/*
 * ocache_disk.c --- disk handling procedures for object caching at KOTETU.
 *      by k-chinen@is.aist-nara.ac.jp, 2000, 2001.
 *
 *  this routines is splitted from xsk.c of kotetu-1.2.
 *
 * $Id$
 */
/*
 * xsk.c --- sk.c + skcache.c
 *      by k-chinen@is.aist-nara.ac.jp, 2000, 2001.
 *
 * $Id$
 */
/*
 * sk.c --- skip list procedures
 *              for KOTETU
 *      by k-chinen@is.aist-nara.ac.jp, 2000
 */
/*
 * skcache.c --- object cache procedures using skip list and its bucket
 *              for KOTETU
 *      by k-chinen@is.aist-nara.ac.jp, 2000
 */

/*

 terminology:
  entry   - object entry. store properties like PCB of TCP/IP
            or super block of filesystem.
  frag    - fragment of object cache. object cache is devided little block
            and stored its list.
            you can access entity of object by drawing its list.
  bucket  - list of entry.  it was controled as skip list.
 
*/
/*
 NOTE:

 hash:
    this program use meny hash to assign object to structures in several level
    and identify among objects.

 on the memory:
    cache consists set of buckets and frag lists in entry.

        bucket[0]
            entry[0]    -> frag[0], frag[1], frag[2] ...
            entry[1]
            entry[2]
            ...

        bucket[1]
            entry[0]
            entry[1]
            entry[2]
            ...

 on the disk:
    cache consists dirctories according to hash.

        aa/bb/cc-dd
        
        aa - first level dir. with first hash (so it is bucket)
        bb - second level dir. with second hash
        cc-dd - file with third and fourth hash.

 */

/*
 * See also:
 *   William Pugh, "skiplist cookbook",
 *   UMIACS-TR-89-72.1, July 1989
 */

/*
 * The algorithm on the original text was described by Pascal like
 * langauage.  Subscription were 1..maxlevel or 1..level .
 * Since this is C program,  This file describes subscription as
 * 0..maxlevel-1 and 0..level-1,
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <pthread.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/uio.h>

#include <dirent.h>

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include <sys/mman.h>

#include <assert.h>
#include <errno.h>


#include "range.h"
#include "msg.h"
#include "trace.h"
#include "misc.h"
#include "mem.h"

#include "ocache.h"
#include "sim_stat.h"
#include "sim_hist.h"
#include "mlist.h"

#ifndef DISK_RETRY
#define DISK_RETRY (13)
#endif

/* object file unit size */
#ifndef OFILE_BLKSIZE
#define OFILE_BLKSIZE       (KB(32))
#endif
#ifndef OFILE_NBLK_FILE
#define OFILE_NBLK_FILE     (509)
#endif
#ifndef OFILE_NBLK_SEEK
#define OFILE_NBLK_SEEK     (OFILE_NBLK_FILE/10)
#endif


#ifndef BODY_FILENAME0
#define BODY_FILENAME0      'B'
#endif
#ifndef BODY_FILENAME 
#define BODY_FILENAME       "BODY"
#endif

#ifndef INDEX_FILENAME0
#define INDEX_FILENAME0     'I'
#endif
#ifndef INDEX_FILENAME  
#define INDEX_FILENAME      "INDEX"
#endif


#ifndef TUP_MAX
/*
#define TUP_MAX (64)
*/
#define TUP_MAX (4096)
#endif

#define H4_DIR_COUNT_MAX    (128)
#define H4_DIR_SIZE_MAX     (128*K(10))
/*
#define H4_DIR_SIZE_MAX     (K(52))
*/

volatile int dir_count_max      = H4_DIR_COUNT_MAX;
volatile off_t dir_size_max     = H4_DIR_SIZE_MAX;
volatile off_t dir_size_hwater  = H4_DIR_SIZE_MAX;
volatile off_t dir_size_lwater  = H4_DIR_SIZE_MAX/2;


/*
 * arround 256, follwing numbers are known as prime numbers.
 *  241 251 257 263 269 271
 *
 */

extern int radix_h1;
extern int radix_h2;
extern int radix_h3;
extern int radix_h4;

extern skcache_entry_t *NIL;
extern int nbucket;
extern bucket_t *skcache_bucket_base;

extern pthread_mutex_t stat_lock;

volatile int count_too_big_object = 0;

volatile int swapin_count       = 0;
volatile int swapin_success     = 0;
volatile int swapout_count      = 0;
volatile int swapout_success    = 0;

volatile int checkdir_dir_count         = 0;
volatile int checkdir_file_read_count   = 0;
volatile int checkdir_file_handle_count = 0;
volatile int checkdir_file_left_count   = 0;
volatile int checkdir_file_delete_count = 0;
volatile int checkdir_file_erase_count  = 0;

volatile int check_dir_ignore_hash_count = 0;

sim_stat_rec    swapin_time;
static long     swapin_time_count[K(1)];
sim_stat_rec    swapout_time;
static long     swapout_time_count[K(1)];

sim_hist_rec    swapin_hist;
static long     swapin_hist_area[K(1)];
sim_hist_rec    swapout_hist;
static long     swapout_hist_area[K(1)];

#ifndef FS_HIST_LEN
#define FS_HIST_LEN (1024)
#endif

volatile int fs_swapin_count[RADIX_H1_MAX];
volatile int fs_swapout_count[RADIX_H1_MAX];

sim_hist_rec    fs_swapout_hist[RADIX_H1_MAX];
static long     fs_swapout_hist_area[RADIX_H1_MAX][FS_HIST_LEN];
sim_hist_rec    fs_swapin_hist[RADIX_H1_MAX];
static long     fs_swapin_hist_area[RADIX_H1_MAX][FS_HIST_LEN];

sim_stat_rec    checkdir_time;
static long     checkdir_time_count[K(2)];
sim_stat_rec    checkdir_time_short;
static long     checkdir_time_short_stat_count[K(2)];

sim_stat_rec    swapout_fail_reason;
static long     swapout_fail_reason_count[K(1)];
sim_stat_rec    swapin_fail_reason;
static long     swapin_fail_reason_count[K(1)];


extern int require_radix_h2;
extern int require_radix_h3;
extern off_t require_dir_size;
#if 0
extern int require_swap_token_val;
extern int require_swap_token_interval;
#endif
extern int require_keeptime_token_val;
extern int require_disk_sweep_interval;



extern pthread_mutex_t config_lock;

extern int skcache_entry_keep_cur;


volatile long long swap_count = 0LL;
volatile long long swap_size = 0LL;

/* limitter */
#if 0
volatile int swapout_token_init = 1000;
volatile int swapout_token  = 1000;
volatile int swapout_token_interval;
#endif
volatile int disk_sweep_interval;

time_t min_date;

extern pthread_mutex_t dir_mutex;
extern int dir_num;
extern int *dir_lock;
extern u_int32_t *dir_sig;
extern off_t *dir_sz;
extern int *dir_count;
extern int *dir_time;
extern int *dir_lack;
extern int dir_conflict;


u_int32_t *dir_bit=NULL;
int *dir_map=NULL;



extern int nindex_dread;
extern int nindex_bdread;
extern int nindex_bdwrite;


TRACE_REGIST_NIC("objcachedisk");


#if 1
/* nanosleep is useful, because it is MT-safe and supports very short time.
 * however, nsec is too short for network I/O.
 * then, we made msleep.  So, 1/1000 sec scale sleep.
 */
/*
inline
*/
static
int
msleep(int x)
{
    struct timespec w;
    long long tw;

    /* convert msec to nsec */
    tw = x * 1000000;
    w.tv_nsec = tw % 1000000000;
    w.tv_sec =  tw / 1000000000;

    nanosleep(&w, NULL);
    
    return 0;
}
#endif



#define OMAP_FUNC(xe)   ((xe)->h3*10000+(xe)->h4)






#ifndef HAVE_DIRENT_T
typedef struct dirent dirent_t;
#endif



#include "ocache_fsmap.c"




char *cache_dir_prefix = "/var/tmp/cache";
int radix_disk = -1;

char *cache_index_prefix = NULL;


/*
 * When it find @ (at-mark) in prefix of cache dirctory,
 * the program recognize using many disk and its pattern.
 */
void
check_disk_spliting()
{
    char *p;
    int v, i;
    char tmp[L_PATH];

    v = -1;
    p = cache_dir_prefix;
    while(*p) {
        if(*p=='@') {
            v = atoi(p+1);
            *p = '\0';
            break;
        }
        p++;
    }

    if(v>0) {
        radix_disk = v;

        Trace(" Spricing: cache directories stored into %d disks\n",
            radix_disk);

#if 0
        for(i=0;i<radix_disk;i++) {
            Trace("   disk# %3d: %s%d\n", i, cache_dir_prefix, i);
        }
#endif
    }

#if 0
    show_bk2fs_map();
    show_fs2bk_map();
#endif
}



/*****
 ***** FILENAME ASSIGNMENT
 *****/

/*
 * kotetu-1.2 and later:
 *  /c0/0/0/INDEX
 *  /c0/0/0/BODY
 *
 * future ?
 *  /c0/0/0-INDEX
 *  /c0/0/0-BODY
 *
 */


inline
char*
assign_file(char *name, char *prefix, skcache_entry_t *e, char *base)
{
    char tmp[L_PATH];

    if(radix_disk<=0) {
        sprintf(name, "%s/%d/%d-%s", prefix, e->h1, e->h2, base);
    }
    else {
#if 0
        sprintf(name, "%s%d/%d/%d-%s",
            prefix, e->h1%radix_disk, e->h1, e->h2, base);
#endif
        sprintf(name, "%s%d/%d/%d-%s",
            prefix, bk2fs[e->h1], e->h1, e->h2, base);
    }

    return name;
}


inline
char*
assign_indexfilename(char *name, skcache_entry_t *e)
{
    assign_file(name, cache_dir_prefix, e, INDEX_FILENAME);
    return name;
}

inline
char*
assign_bodyfilename(char *name, skcache_entry_t *e)
{
    assign_file(name, cache_index_prefix, e, BODY_FILENAME);
    return name;
}





pthread_mutex_t fd_assign_lock = PTHREAD_MUTEX_INITIALIZER;
int *ifd_pool;
int *bfd_pool;


/*****
 *****
 ***** position caching and seek faking.
 *****
 *****/



off_t *ips_pool;
off_t *bps_pool;


/*
 * return:
 *      0   right
 *      !0  strange
 */

off_t
raw_pos_verify(int fd, off_t pos, char *fname, int line)
{
    off_t apos;
    int c;

    if(fd<0) {
        Error("pv: fail fd# %d, pos %lld : %s %d",
            fd, (long long)pos, fname, line);
        return 0;
    }
    if(pos<(off_t)0) {
#if 0
        Error("pv: fail fd# %d, pos %lld : %s %d",
            fd, (long long)pos, fname, line);
#endif
        return 0;
    }

    c = 0;
try:
    c++;
    apos = lseek(fd, 0, SEEK_CUR);
    if(apos<(off_t)0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try;
        }
        Error("pv: fd# %d, error %d : %s %d",
            fd, errno, fname, line);
    }
    
    if(apos==pos) {
#if 0
        Trace("pv: fd# %d, pos view %lld, actual %lld ; equal\n",
            fd, (long long)pos, (long long)apos);
#endif
        return 0;
    }
    else {
        Error("pv: fd# %d, pos view %lld, actual %lld ; diff %lld : %s %d",
            fd, (long long)pos, (long long)apos, (long long) apos-pos,
            fname, line);
        return apos-pos;
    }
}


#ifdef TRACE_POS
#define POSCHECKF(fpos) \
    { if(fpos<0 && fpos>=dir_num) { \
        Error("Iseek: fpos %d out of range [%d..%d) : %s %d\n", \
            fpos, 0, dir_num, __FILE__, __LINE__); \
      } \
    }
#define POSCHECKI(vpos) \
    { if(vpos<0L || vpos>(OFILE_NBLK_FILE*ESIZE)) { \
        Error("pos-range: index overrun, pos %lldB <%ld> : %s %d", \
            (long long)vpos, (long)(vpos/ESIZE), \
            __FILE__, __LINE__); \
        } else { \
        } \
    }
#define POSCHECKP(vpos) \
    { if(vpos<0L || vpos>(OFILE_NBLK_FILE*OFILE_BLKSIZE)) { \
        Error("pos-range: pack overrun, pos %lldB <%ld>: %s %d", \
            (long long)vpos, (long)(vpos/OFILE_BLKSIZE), \
            __FILE__, __LINE__); \
        } else { \
        } \
    }

#else /* TRACE_POS */

#define POSCHECKF(x)        (void)1
#define POSCHECKI(x)        (void)1
#define POSCHECKP(x)        (void)1

#endif /* TRACE_POS */


#ifdef TRACE_POS2

#define POSVERIFY(xxf,xxp)  raw_pos_verify(xxf, xxp, __FILE__, __LINE__)
#define LENVERIFY(xxf)      (void)1

#else

#define POSVERIFY(xxf,xxp)  (void)1
#define LENVERIFY(xxf)      (void)1

#endif /* TRACE_POS2 */


/*
 * seeking for packed file.
 */
inline
static
off_t 
Iseek(int fpos, int sfd, off_t pos)
{
    off_t dst;
    off_t dpos;
    int ddir;
    int c;

    POSCHECKF(fpos);
    POSCHECKI(pos);

    if(fpos<0 && fpos>=dir_num) {
        Error("Iseek: fpos %d out of range [%d..%d)\n",
            fpos, 0, dir_num);
    } 
    if(pos<(off_t)0) {
        goto out;
    }

#ifndef NO_SEEKFAKE
    if(ips_pool[fpos]<0)  { ddir = SEEK_SET;    dpos = pos; }
    else {
        if(pos == ips_pool[fpos]) {
#ifdef TRACE_POS2
    Trace("Iseek: dir# %d, fd# %d, pos %lld, last %lld => no-move\n",
        fpos, sfd, (long long)pos, (long long)ips_pool[fpos]);
#endif
            dst = pos;
            goto no_move;
        }
#if 0
#endif
        if(pos==(off_t)0) { ddir = SEEK_SET;    dpos = 0; }
        else {              ddir = SEEK_CUR;    dpos = pos - ips_pool[fpos];}
    }
#else
    ips_pool[fpos] = -1;
    ddir = SEEK_SET;    dpos = pos;
#endif /* FAKE_SEEK */

#ifdef TRACE_POS2
    Trace("Iseek: start dir# %d, fd# %d, pos %d %lld, last %lld => lseek %lld %d\n",
        fpos, sfd, (long)(pos/ESIZE), (long long)pos,
        (long long)ips_pool[fpos], (long long)dpos, ddir);
#endif

    c = 0;
try_seek:
    c++;
    dst = lseek(sfd, dpos, ddir);
    if(dst<(off_t)0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_seek;
        }
        Error(
"Iseek: fail dir# %d, fd# %d, cannot seek <%lldB dir %d> %lldB on index file : try %d errno %d",
            fpos, sfd, (long long)dpos, ddir, (long long)pos, c, errno);
        dst = ips_pool[fpos] = -1;
        goto out;
    }
    ips_pool[fpos] = dst;

no_move:
out:
    POSCHECKI(dst);
    POSVERIFY(sfd, dst);

    return dst;
}


inline
static
off_t 
Pseek(int fpos, int sfd, off_t pos)
{
    off_t dst;
    off_t dpos;
    int ddir;
    int c;

    POSCHECKF(fpos);
    POSCHECKP(pos);

    if(fpos<0 && fpos>=dir_num) {
        Error("Pseek: fpos %d out of range [%d..%d)\n",
            fpos, 0, dir_num);
    }

#ifndef NO_SEEKFAKE
    if(bps_pool[fpos]<0) {  ddir = SEEK_SET;    dpos = pos; }
    else {
        if(pos == bps_pool[fpos]) {
#ifdef TRACE_POS
    Trace("Pseek: dir# %d, fd# %d, pos %lld, last %lld => no-move\n",
        fpos, sfd, (long long)pos, (long long)bps_pool[fpos]);
#endif
            dst = pos;
            goto nomove;
        }
        if(pos==(off_t)0) { ddir = SEEK_SET;    dpos = 0; }
        else {              ddir = SEEK_CUR;    dpos = pos - bps_pool[fpos]; }
    }
#else
    bps_pool[fpos] = -1;
    ddir = SEEK_SET;    dpos = pos;
#endif

#ifdef TRACE_POS2
    Trace("Pseek: start dir# %d, fd# %d, goto pos %d %lld, last %lld => lseek %lld %d\n",
        fpos, sfd, (long)(pos/OFILE_BLKSIZE), (long long)pos,
        (long long)bps_pool[fpos], (long long)dpos, ddir);
#endif


    c = 0;
try_seek:
    c++;
    dst = lseek(sfd, dpos, ddir);
    if(dst<(off_t)0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_seek;
        }
        Error("Pseek: fail dir# %d, fd# %d, cannot seek <%lldB dir %d> %lldB on index file : try %d errno %d",
            fpos, sfd,
            (long long)dpos, ddir, (long long)pos, c, errno);

        dst = bps_pool[fpos] = -1;
        goto out;
    }
    bps_pool[fpos] = dst;


nomove:
out:
    POSCHECKP(dst);
    POSVERIFY(sfd, dst);

    return dst;
}


int
Iread(int fpos, int fd, void* buf, int len)
{
    int ret;
    int raw_ret;
    int c;
    off_t start_pos;

    if(fpos<0 && fpos>=dir_num) {
        Error("Iread: fpos %d out of range [%d..%d)\n",
            fpos, 0, dir_num);
    }
    POSCHECKF(fpos);
    start_pos = ips_pool[fpos];
    c = 0;
try:
    c++;
    raw_ret = read(fd, buf, len);
    if(raw_ret<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            Error("Iread: fail read(); dir# %d, fd %d, pos %lld, len %d (%d)\n",
                fpos, fd, start_pos, len, errno);
            goto try;
        }
        ips_pool[fpos] = -1;
    }
    if(raw_ret>0) {
        ips_pool[fpos] += raw_ret;
#ifdef TRACE_POS2
        Trace("Iread: dir# %d, fd# %d, read %ld/%ld, moveto %lld\n",
            fpos, fd, (long)raw_ret, (long)len, (long long)ips_pool[fpos]);
#endif
    }

    return raw_ret;
}

int
Ireadv(int fpos, int fd, struct iovec *vb, int vc)
{
    int ret;
    int raw_ret;
    int c;
    off_t start_pos;

    if(fpos<0 && fpos>=dir_num) {
        Error("Ireadv: fpos %d out of range [%d..%d)\n",
            fpos, 0, dir_num);
    }
    POSCHECKF(fpos);
    start_pos = ips_pool[fpos];
    c = 0;
try:
    c++;
    raw_ret = readv(fd, vb, vc);
    if(raw_ret<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            Error("Iread: fail readv(); dir# %d, fd %d, pos %lld, vc %d (%d)\n",
                fpos, fd, start_pos, vc, errno);
            goto try;
        }
        ips_pool[fpos] = -1;
    }
    if(raw_ret>0) {
        ips_pool[fpos] += raw_ret;
#ifdef TRACE_POS2
        Trace("Ireadv: dir# %d, fd# %d, read %ld, moveto %lld\n",
            fpos, fd, (long)raw_ret, (long long)ips_pool[fpos]);
#endif
    }

    return raw_ret;
}

int
Iwrite(int fpos, int fd, void* buf, int len)
{
    int ret;
    int raw_ret;
    int c;
    off_t start_pos;

    if(fpos<0 && fpos>=dir_num) {
        Error("Iwrite: fpos %d out of range [%d..%d)\n",
            fpos, 0, dir_num);
    }
    POSCHECKF(fpos);
    start_pos = ips_pool[fpos];
    c = 0;
try:
    c++;
    raw_ret = write(fd, buf, len);
    if(raw_ret<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            /*
            Error("Iwrite: fail write() index; dir# %d, fd %d, pos %lld, len %d (%d)\n",
                fpos, fd, start_pos, len, errno);
            */
            goto try;
        }
        ips_pool[fpos] = -1;
        Error("Iwrite: fail write() index; dir# %d, fd %d, pos %lld, len %d (%d)\n",
            fpos, fd, start_pos, len, errno);
    }
    if(raw_ret>0) {
        ips_pool[fpos] += raw_ret;
#ifdef TRACE_POS2
        Trace("Iwrite: dir# %d, fd# %d, write %ld/%ld, moveto %lld\n",
            fpos, fd, (long)raw_ret, (long)len, (long long)ips_pool[fpos]);
#endif
    }

    return raw_ret;
}

int
Iwritev(int fpos, int fd, struct iovec *vb, int vc)
{
    int ret;
    int raw_ret;
    int c;
    off_t start_pos;

    if(fpos<0 && fpos>=dir_num) {
        Error("Iwritev: fpos %d out of range [%d..%d)\n",
            fpos, 0, dir_num);
    }
    POSCHECKF(fpos);
    start_pos = ips_pool[fpos];
    c = 0;
try:
    c++;
    raw_ret = writev(fd, vb, vc);
    if(raw_ret<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            /*
            Error("Iwritev: fail writev() index; dir# %d, fd %d, pos %lld, vc %d (%d)\n",
                fpos, fd, start_pos, vc, errno);
            */
            goto try;
        }
        ips_pool[fpos] = -1;
        Error("Iwritev: fail writev() index; dir# %d, fd %d, pos %lld, vd %d (%d)\n",
            fpos, fd, start_pos, vc, errno);
    }
    if(raw_ret>0) {
        ips_pool[fpos] += raw_ret;
#ifdef TRACE_POS2
        Trace("Iwritev: dir# %d, fd# %d, write %ld, moveto %lld\n",
            fpos, fd, (long)raw_ret, (long long)ips_pool[fpos]);
#endif
    }

    return raw_ret;
}


int
Pread(int fpos, int fd, void* buf, int len)
{
    int ret;
    int raw_ret;
    int c;
    off_t start_pos;

    if(fpos<0 && fpos>=dir_num) {
        Error("Pread: fpos %d out of range [%d..%d)\n",
            fpos, 0, dir_num);
    }
    POSCHECKF(fpos);
    start_pos = bps_pool[fpos];
    c = 0;
try:
    c++;
    raw_ret = read(fd, buf, len);
    if(raw_ret<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            /*
            Error("Pread: fail read() data; dir# %d, fd %d, pos %lld, len %d (%d)\n",
                fpos, fd, start_pos, len, errno);
            */
            goto try;
        }
        bps_pool[fpos] = -1;
        Error("Pread: fail read() data; dir# %d, fd %d, pos %lld, len %d (%d)\n",
            fpos, fd, start_pos, len, errno);
    }
    if(raw_ret>0) {
        bps_pool[fpos] += raw_ret;
#ifdef TRACE_POS2
        Trace("Pread: dir# %d, fd# %d, read %ld/%ld, moveto %lld\n",
            fpos, fd, (long)raw_ret, (long)len, (long long)bps_pool[fpos]);
#endif
    }

    return raw_ret;
}

int
Preadv(int fpos, int fd, struct iovec *vb, int vc)
{
    int ret;
    int raw_ret;
    int c;
    off_t start_pos;

    if(fpos<0 && fpos>=dir_num) {
        Error("Preadv: fpos %d out of range [%d..%d)\n",
            fpos, 0, dir_num);
    }
    POSCHECKF(fpos);
    start_pos = bps_pool[fpos];
    c = 0;
try:
    c++;
    raw_ret = readv(fd, vb, vc);
    if(raw_ret<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            /*
            Error("Preadv: fail readv() data; dir# %d, fd %d, pos %lld, vc %d (%d)\n",
                fpos, fd, start_pos, vc, errno);
            */
            goto try;
        }
        bps_pool[fpos] = -1;
        Error("Preadv: fail readv() data; dir# %d, fd %d, pos %lld, vc %d (%d)\n",
            fpos, fd, (long long)start_pos, vc, errno);
    }
    if(raw_ret>0) {
        bps_pool[fpos] += raw_ret;
#ifdef TRACE_POS2
        Trace("Preadv: dir# %d, fd# %d, read %ld, moveto %lld\n",
            fpos, fd, (long)raw_ret, (long long)bps_pool[fpos]);
#endif
    }

    return raw_ret;
}


int
Pwrite(int fpos, int fd, void* buf, int len)
{
    int ret;
    int raw_ret;
    int c;
    off_t start_pos;

    if(fpos<0 && fpos>=dir_num) {
        Error("Pwrite: fpos %d out of range [%d..%d)\n",
            fpos, 0, dir_num);
    }
    POSCHECKF(fpos);
    start_pos = bps_pool[fpos];
    c = 0;
try:
    c++;
    raw_ret = write(fd, buf, len);
    if(raw_ret<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            /*
            Error("Pwrite: fail write() data; dir# %d, fd %d, pos %lld, len %d (%d)\n",
                fpos, fd, start_pos, len, errno);
            */
            goto try;
        }
        bps_pool[fpos] = -1;
        Error("Pwrite: fail write() data; dir# %d, fd %d, pos %lld, len %d (%d)\n",
                fpos, fd, (long long)(long long)(long long)(long long)(long long)(long long)(long long)(long long)start_pos, len, errno);
    }
    if(raw_ret>0) {
        bps_pool[fpos] += raw_ret;
#ifdef TRACE_POS2
        Trace("Pwrite: dir# %d, fd# %d, write %ld/%ld, moveto %lld\n",
            fpos, fd, (long)raw_ret, (long)len, (long long)bps_pool[fpos]);
#endif
    }

    return raw_ret;
}

int
Pwritev(int fpos, int fd, struct iovec *vb, int vc)
{
    int ret;
    int raw_ret;
    int c;
    off_t start_pos;

    if(fpos<0 && fpos>=dir_num) {
        Error("Pwritev: fpos %d out of range [%d..%d)\n",
            fpos, 0, dir_num);
    }
    POSCHECKF(fpos);
    start_pos = bps_pool[fpos];
    c = 0;
try:
    c++;
    raw_ret = writev(fd, vb, vc);
    if(raw_ret<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            Error("Pwritev: fail writev() data; dir# %d, fd %d, pos %lld, vc %d (%d)\n",
                fpos, fd, start_pos, vc, errno);
            goto try;
        }
        bps_pool[fpos] = -1;
        Error("Pwritev: fail writev() data; dir# %d, fd %d, pos %lld, vc %d (%d)\n",
            fpos, fd, (long long)start_pos, vc, errno);
    }
    if(raw_ret>0) {
        bps_pool[fpos] += raw_ret;
#ifdef TRACE_POS2
        Trace("Pwritev: dir# %d, fd# %d, write %ld, moveto %lld\n",
            fpos, fd, (long)raw_ret, (long long)bps_pool[fpos]);
#endif
    }

    return raw_ret;
}






/*
 * return:
 *      0   success
 *      !0  fail and indicate reason
 *
 */
int
ifd_open(int *rfd, bucket_t *bk, skcache_entry_t *e)
{
    int ret;
    int sfd;
    char name[L_PATH];
    char *namechk;
    int c;
    off_t pos, dst;
    skcache_entry_t _xe;
    skcache_entry_t *xe;
    int fpos;
    struct stat sbuf;
    int chk;

    ret = 0;
    *rfd = -1;

#ifdef TRACE_POS
    Trace("ifd_open: [%03d:%03d:%03d:%04d] len %ld/%ld\n",
        e->h1, e->h2, e->h3, e->h4, (long)e->totallength, (long)OFILE_BLKSIZE);
#endif

    /* only use h1 and h2 */
    xe = &_xe;
    *xe = *e;
    xe->h3 = 0; 
    xe->h4 = 0; 

    fpos = FDPOOL_POS_E(xe);

#ifdef TRACE_POS2
    Trace("ifd_open: [%03d:%03d:%03d:%04d] len %ld/%ld, pos %ld\n",
        e->h1, e->h2, e->h3, e->h4,
        (long)e->totallength, (long)OFILE_BLKSIZE,
        (long)fpos);
#endif

    MUTEX_LOCK(&fd_assign_lock);

    sfd = ifd_pool[fpos];
    if(sfd>=0) {
        *rfd = sfd;
#ifdef TRACE_POS
Trace("ifd_open: cached %4d %d,%d *rfd %d, return %d\n",
    fpos, xe->h1, xe->h2, *rfd, ret);
#endif
        goto out;
    }


    namechk = assign_indexfilename(name, xe);

    c = 0;
try_open:
    c++;
    sfd = open(name, O_RDWR | O_CREAT, 0666);
    if(sfd<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_open;
        }
#if 0
        if(errno==ENOENT && c<DISK_RETRY) {
            build_dir_bynumber(xe->h1, xe->h2, xe->h3);
            goto try_open;
        }
#endif
        if(errno==ENOSPC) {
            Error("ifd_open: disk full");
            ret = XSK_DISKFULL;
            goto fail_open;
        }
        Error("ifd_open: cannot open swap file '%s' %d (%d)", name, c, errno);

        ret = XSK_INDEX_EOPEN;
        goto fail_open;
    }

#if 1
    chk = stat(name, &sbuf);
    if(chk) {
        Error("fail stat() for INDEX '%s' (%d)", name, errno);
    }
    else {
        if(sbuf.st_size != (off_t) OFILE_NBLK_FILE*ESIZE) {
            Trace("INDEX %-16s fd# %d, ignore size %9lldB %lds ; expect %9lldB %6ldB x %lds\n",
                name, sfd,
                (long long)sbuf.st_size,
                (long)((long long)sbuf.st_size/ESIZE),
                (long long) OFILE_NBLK_FILE*ESIZE,
                (long)ESIZE, (long)OFILE_NBLK_FILE);
        }
    }
#endif

    ifd_pool[fpos] = sfd;
    *rfd = sfd;

#ifdef TRACE_POS2
Trace("ifd_open: new    %4d %d,%d *rfd %d, return %d\n",
    fpos, xe->h1, xe->h2, *rfd, ret);
#endif

    goto out;

fail_open:
    (void)1;

out:

    MUTEX_UNLOCK(&fd_assign_lock);

#ifdef TRACE_POS
Trace("ifd_open: oid# %d, *rfd %d, return %d\n", xe->oid, *rfd, ret);
#endif

    return ret;
}

int
bfd_open(int *rfd, bucket_t *bk, skcache_entry_t *e)
{
    int ret;
    int sfd;
    char name[L_PATH];
    char *namechk;
    int c;
    off_t pos, dst;
    skcache_entry_t _xe;
    skcache_entry_t *xe;
    int fpos;
    struct stat sbuf;
    int chk;

    ret = 0;
    *rfd = -1;

#ifdef TRACE_POS
    Trace("bfd_open: [%03d:%03d:%03d:%04d] len %ld/%ld\n",
        e->h1, e->h2, e->h3, e->h4, (long)e->totallength, (long)OFILE_BLKSIZE);
#endif

    /* only use h1 and h2 */
    xe = &_xe;
    *xe = *e;
    xe->h3 = 0; 
    xe->h4 = 0; 

    fpos = FDPOOL_POS_E(xe);

#ifdef TRACE_POS2
    Trace("bfd_open: [%03d:%03d:%03d:%04d] len %ld/%ld, pos %ld\n",
        e->h1, e->h2, e->h3, e->h4,
        (long)e->totallength, (long)OFILE_BLKSIZE,
        (long)fpos);
#endif

    MUTEX_LOCK(&fd_assign_lock);

    sfd = bfd_pool[fpos];
    if(sfd>=0) {
        *rfd = sfd;
#if 0
Trace("bfd_open: cached %d,%d *rfd %d, return %d\n",
    xe->h1, xe->h2, *rfd, ret);
#endif
        goto out;
    }


    namechk = assign_bodyfilename(name, xe);

    c = 0;
try_open:
    c++;
    sfd = open(name, O_RDWR | O_CREAT, 0666);
    if(sfd<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_open;
        }
#if 0
        if(errno==ENOENT && c<DISK_RETRY) {
            build_dir_bynumber(xe->h1, xe->h2, xe->h3);
            goto try_open;
        }
#endif
        if(errno==ENOSPC) {
            Error("bfd_open: disk full on open()");
            ret = XSK_DISKFULL;
            goto fail_open;
        }
        Error("bfd_open: cannot open swap file '%s' %d (%d)", name, c, errno);

        ret = XSK_BODY_EOPEN;
        goto fail_open;
    }

#if 1
    chk = stat(name, &sbuf);
    if(chk) {
        Error("fail stat() for BODY '%s' (%d)", name, errno);
    }
    else {
        if(sbuf.st_size != (off_t) OFILE_NBLK_FILE*OFILE_BLKSIZE) {
            Trace("BODY  %-16s fd# %d, ignore size %9lldB %lds ; expect %9lldB %6ldB x %lds\n",
                name, sfd,
                (long long)sbuf.st_size,
                (long)((long long)sbuf.st_size/OFILE_BLKSIZE),
                (long long) OFILE_NBLK_FILE*OFILE_BLKSIZE,
                (long)OFILE_BLKSIZE, (long)OFILE_NBLK_FILE);
        }
    }
#endif

    bfd_pool[fpos] = sfd;
    *rfd = sfd;

#ifdef TRACE_POS2
Trace("bfd_open: new    %d,%d sfd %d, *rfd %d, return %d\n",
    xe->h1, xe->h2, sfd, *rfd, ret);
#endif

    goto out;

fail_open:
    (void)1;

out:

    MUTEX_UNLOCK(&fd_assign_lock);

#ifdef TRACE_POS
Trace("bfd_open: oid# %d, *rfd %d, return %d\n", xe->oid, *rfd, ret);
#endif

    return ret;
}

/* this routine never called */
int
bfd_close(int *rfd)
{
#if 0
    if(*rfd>=0) {
        close(*rfd);
        *rfd = -1;
    }
#endif
}



void
dbit_mkhumanstr(u_int32_t *s, char *line)
{
    int i;
    u_int32_t m, v;
    char *p;

    p = line;
    for(i=0;i<DBIT_BIT_WIDTH;i++) {
        m = 1U << i%(SIZE_OF_UINT*8);
        v = s[i/(SIZE_OF_UINT*8)] & m;
        *p++ = (v ? '1' : '0');
    }
    *p = '\0';
}

void
dbit_dump(u_int32_t *s)
{
    char line[DBIT_BIT_WIDTH+2];    /* keep more 2 chars for \n + \0 */

    dbit_mkhumanstr(s, line);
    Trace("%s\n", line);
}

void
dbit_show(u_int32_t *s)
{
    char line[DBIT_BIT_WIDTH+2];    /* keep more 2 chars for \n + \0 */
    int i, c;

    dbit_mkhumanstr(s, line);
    c = 0;
    for(i=0;i<DBIT_BIT_WIDTH;i++) {
        if(DBIT_TMP_ISSET(s, i)) {
            c++;
        }
    }
    Trace("addr %p count %d, bits %s\n", s, c, line);
}


void
dbit_set(u_int32_t *s, int i)
{
    u_int32_t m;
    m = 1UL << i%(SIZE_OF_UINT*8);
    s[i/(SIZE_OF_UINT*8)] |= m;
}

int
dbit_isset(u_int32_t *s, int i)
{
    u_int32_t m;
    m = 1UL << i%(SIZE_OF_UINT*8);
    return (s[i/(SIZE_OF_UINT*8)] & m) ? 1 : 0;
}

int
dbit_allset(u_int32_t *s)
{
    int i;
    u_int32_t u;

    for(i=0;i<DBIT_ELEM_WIDTH;i++) {
        u = (u_int32_t)-1L; /* all bits were set */
        s[i] = u;
    }
}

int
dbit_allclear(u_int32_t *s)
{
    int i;
    u_int32_t u;

    for(i=0;i<DBIT_ELEM_WIDTH;i++) {
        u = (u_int32_t)0UL;
        s[i] = u;
    }
}










/*
#define ofile_offset(e) ((e->h3+e->h4)%OFILE_NBLK_FILE)
*/

inline
int
old_ofile_offset(skcache_entry_t *e)
{
#if 0
    return (e->h3+e->h4)%OFILE_NBLK_FILE;
#endif
    return ((e->h3*e->h4)%(OFILE_NBLK_FILE/OFILE_NBLK_OBJ))*OFILE_NBLK_OBJ;
}

/*
 it is defined in ocache.h
#define ofile_offset(e) (((e->h3*e->h4)%(OFILE_NBLK_FILE/OFILE_NBLK_OBJ))*OFILE_NBLK_OBJ)
*/


#if 0

int
dump_ifile(int ifd, int start_pos)
{
    int i;
    int n;
    skcache_entry_t tmp_entry;
    off_t pos;
    int nr;
    int diff;

    n = start_pos;
    for(i=0;i<OFILE_NBLK_SEEK;i++) {
        pos = n*ESIZE;
        lseek(ifd, pos, SEEK_SET);

        nr = read(ifd, &tmp_entry, ESIZE);

        diff = tmp_entry.issue_time - min_date;
    Trace("  :: fd# %d, try %3d: pos %4d %12lldB : magic(hex) %02x %02x, oid# %d, [%d] diff %d\n",
            ifd, i, n, (long long)pos,
            (unsigned char)tmp_entry.magic[0],
            (unsigned char)tmp_entry.magic[1],
            tmp_entry.oid, tmp_entry.h4, diff
        );

        n = (n+1)%OFILE_NBLK_FILE;
    }

    return 0;
}

#endif

#define read(x1,x2,x3)  "error" __FILE__ __LINE__
#define write(x1,x2,x3) "error" __FILE__ __LINE__
#define readv(x1,x2,x3) "error" __FILE__ __LINE__
#define writev(x1,x2,x3)    "error" __FILE__ __LINE__



/* 
 * XXX 'first match'
 *
 * seek bitmap on memory
 *
 *
 * return:
 *      -1  not found or error
 *      0>  found and its position
 *              for example, the fuction returns 3.
 *              blocks between 3 to 3+want-1 are free.
 *              the caller should use them for new swap.
 */
static
int
_psf_bit_firstmatch(int sig_pos, int want, int npos, int verbose)
{
    int ii, i;
    int v;
    int match;
    int first;
    int try;
    int ret;

    ret = -1;

#if 0
    for(i=npos;i<MIN(npos+OFILE_NBLK_SEEK, OFILE_NBLK_FILE);i++) {
        v = DBIT_ISSET(sig_pos, i);
        Trace("psf: %3d %4d bit %d \n", xid, i, v);
    }
#endif

    try = OFILE_NBLK_SEEK;
    match = 0;
    first = -1;
    for(ii=npos;ii<npos+OFILE_NBLK_SEEK;ii++)
    { 
        i = ii%OFILE_NBLK_FILE;

        if(i==0) {
            match = 0;
            first = -1;
        }

        v = DBIT_ISSET(sig_pos, i);

        if(v==0) {
            if(first<0) {
                first = i;
            }
            match++;
        }
        else {
            match = 0;
            first = -1;
        }

 if(verbose) {
Trace("  :F dir# %d, pos %4d %4d, try %3d: v %d match %2d first %4d\n",
        sig_pos, i, ii, try, v, match, first);
 }

        if(match>=want) {
            break;
        }

        try--;
    }

    if(match>=want) {
        ret = first;
    }
    else {
        ret = -1;
    }
 if(verbose) {
Trace("  : dir# %d, ; return %d\n", sig_pos, ret);
 }

    return ret;
}

#ifndef BMRADIX 
#define BMRADIX (53)
#endif
#define DFC ((e->h1)%BMRADIX==1)

/* print debug information of several file by select 1/53 of all sessions
 * DFC is masking for such selection.
 */


/*
 * return:
 *  success     0
 *  error       <0
 *
 */
int
bfd_seekfree(int sfd, bucket_t *bk, skcache_entry_t *e, int *rpos)
{
    int ret;
    int c;
    off_t pos, dst;
    int fpos;
    int npos;
    int nr;
    skcache_entry_t *tmp_entry;
    int match = 0;
    int try;
    off_t vpos;
    time_t diff;
    skcache_entry_t chunk[HARD_NINDEX_DREAD];
    int i;
    int nwid, wid;
    int v;
    int first;
    int want;

    int xid; /* loop id */
    int sig_pos;

    int tmp_pos;

    ret = 0;
    *rpos = (off_t) -1;

    xid = rand()%1000;  /* make temporary loop id */

    
    if(e==NULL) {
    }

    fpos = FDPOOL_POS_E(e);
    sig_pos = DSIG_POS(e->h1, e->h2, e->h3);
    want = NBCEIL(e->totallength, OFILE_BLKSIZE);

    npos = ofile_offset(e);
    pos = (off_t)npos * ESIZE;
    if(pos<0 || npos<0) {
        Error("psf: 1 Ignore position block# %d %lldB; e h3 %d h4 %d; slot %d",
            npos, (long long)pos, e->h3, e->h4, OFILE_NBLK_FILE);
        ret = XSK_BODY_EPOS;
        goto out;
    }

#if 0
    Trace("psf: 0 [%03d:%03d:%03d:%04d] len %ld/%ld\n",
        e->h1, e->h2, e->h3, e->h4, (long)e->totallength, (long)OFILE_BLKSIZE);
#endif

#if 0
 if(DFC) {
Trace("psf: 2 sfd# %d, [%03d:%03d:%03d:%04d] x %6ld => pos %8d/%d %12lldB (%4ldMB+%7ld)\n",
    sfd,
    e->h1, e->h2, e->h3, e->h4, OFILE_BLKSIZE,
    npos, OFILE_NBLK_FILE,
    (long long)pos, (long)dM(pos), (long)mM(pos) );
 }
#endif

 if(DFC) {
Trace("psf: 2 dir# %d, sfd# %d, %d:%d:%d:%d from pos %d with %d blocks ; count %d\n",
    sig_pos, sfd, e->h1, e->h2, e->h3, e->h4, npos, want, dir_count[sig_pos]);
 }



#if 1
    if(dir_count[sig_pos]<0) {
#if 0
        Trace("psf: %3d don't ref. dir_bit\n", xid);
#endif

        /* XXX, currently this disk seeking routine does not support
           multi block seeking */

#if 0
        if(want>1) {
            goto not_found;
        }
#endif
#if 0
        if(want>1) {
            Trace("  oid# %d, want %d\n", e->oid, want);
        }
#endif

    }
    else {

        tmp_pos = _psf_bit_firstmatch(sig_pos, want, npos, DFC);
#if 0
        Trace("result by first-match %d\n", tmp_pos);
#endif

        if(tmp_pos<0) {
            goto not_found;
        }
        else {
            npos = tmp_pos;
            goto found_by_membits;
        }
    }
#endif


    if(pos<(off_t)0L) {
        Error("psf: ignore position block# %d %lldB before lseek()", 
            npos, (long long)pos);

        ret = XSK_BODY_EPOS;
        goto out;
    }

#if 0
    if(POSVERIFY(sfd, ips_pool[sig_pos])) {
        ips_pool[sig_pos] = -1;
    }
#endif

    c = 0;
try_seek:
    c++;
#if 0
    dst = lseek(sfd, pos, SEEK_SET);
#endif
    dst = Iseek(sig_pos, sfd, pos);
    if(dst<(off_t)0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_seek;
        }
Error("psf: cannot seek %d %lldB on index file fd# %d :%d (%d)",
            npos, (long long)pos, sfd, c, errno);

        ret = XSK_BODY_EPOS;
        goto out;
    }
    vpos = dst;
    POSVERIFY(sfd, vpos);


    try = OFILE_NBLK_SEEK;
    match = 0;
    first = -1;
    do {
        pos = (off_t)npos * ESIZE;
        nwid = (MIN(npos+ MIN(nindex_dread, OFILE_NBLK_SEEK), OFILE_NBLK_FILE) -  npos);
        wid = ESIZE*nwid;

        /* block read */

 if(DFC) {
Trace("psf: 4 sfd# %d, pos %d %lld, addr %p, wid %d %d => %d/%d %lld/%lld\n",
    sfd, npos, (long long)pos, chunk, nwid, wid, npos+nwid, OFILE_NBLK_FILE,
    (long long)(pos+wid), (long long)(OFILE_NBLK_FILE*ESIZE) );
 }

        c = 0;
try_read:
        c++;

#if 0
        nr = read(sfd, chunk, wid);
#endif
        nr = Iread(sig_pos, sfd, chunk, wid);
        if(nr<0) {
            if(errno==EINTR && c < DISK_RETRY) {
                goto try_read;
            }
            Error("psf: cannot read fd# %d: %d (%d)",
                sfd, c, errno);
            break;
        }
        vpos += nr;
        POSVERIFY(sfd, vpos);
        nwid = nr/ESIZE;

        for(i=0;i<nwid && try;i++) {
            pos = (off_t)npos * ESIZE;
            tmp_entry = &chunk[i];

            if(npos==0) {
                match = 0;
                first = -1;
            }

            if(tmp_entry->magic[0]!=MAGIC_0 || tmp_entry->magic[1]!=MAGIC_1
                || tmp_entry->oid == 0 || tmp_entry->issue_time < min_date) {

                match++;
                if(first<0) {
                    first = npos;
                }
            }
            else {
                match = 0;
                first = -1;
            }

#ifdef TRACE_INDEX
#endif
 if(DFC) {
        diff = tmp_entry->issue_time - min_date;
Trace("  :f dir# %d, fd# %d, try %3d: pos %4d %12lldB : vpos %12lldB : magic(hex) %02x %02x, oid# %d, diff %d; match %d, first %d\n",
        sig_pos, sfd, try, npos, (long long)pos, (long long) vpos,
        (unsigned char)tmp_entry->magic[0], (unsigned char)tmp_entry->magic[1],
        tmp_entry->oid, diff, match, first
        );
 }

            if(match>=want) {
                break;
            }
    
            npos = (npos+1)% OFILE_NBLK_FILE;
            try--;

            if(npos==0) {
                break;
            }
        }
        if(match>=want) {
            break;
        }
        
        if(npos==0) {
#ifdef TRACE_INDEX
            Trace("  : rewind sfd# %d,\n", sfd);
#endif
            c = 0;
try_rewind:
            c++;
#if 0
            dst = lseek(sfd, 0, SEEK_SET);
#endif
            dst = Iseek(sig_pos, sfd, 0);
            if(dst<(off_t)0) {
                if(errno==EINTR && c<DISK_RETRY) {
                    goto try_rewind;
                }
    Error("psf: cannot rewind %lld on index file fd# %d :%d (%d)",
        (long long)pos, sfd, c, errno);
            }
            vpos = dst;
            POSVERIFY(sfd, vpos);
        
            match = 0;
            first = -1;
        }
    } while (match==0 && try>0);


    /* 
     * find by disk seeking
     *  - move file pointer to first block on INDEX file
     *  - set return code and position
     */

    if(match && first>=0) {
#if 0
    Trace("psf: %3d found in disk; block# %d + %d\n", xid, npos, want);
#endif

        c = 0;
try_back:
        c++;
        npos = first;
        pos = first*ESIZE;
#if 0
        dst = lseek(sfd, - (vpos - pos), SEEK_CUR);
#endif
        dst = Iseek(sig_pos, sfd, pos);
        if(dst<(off_t)0) {
            if(errno==EINTR && c<DISK_RETRY) {
                goto try_back;
            }
    Error("psf: cannot back %lld on index file fd# %d :%d (%d)",
        (long long)pos, sfd, c, errno);

            ret = XSK_BODY_EROLL;
            *rpos = npos;
            goto out;
        }
        vpos = dst;
        POSVERIFY(sfd, vpos);

#if 0
        ret = npos;
#endif
        ret = 0;
        *rpos = npos;

        goto found;
    }
    else {
#if 0
Trace("psf: 1 sfd# %d, not found (want %d), [%d:%d:%d:%d] fpos %d, npos %d\n",
        sfd, want, e->h1, e->h2, e->h3, e->h4, FDPOOL_POS_E(e), ofile_offset(e));
#endif
        goto not_found;
    }

    goto out;


    /* 
     * find by bitmask on memory seeking
     *  - move file pointer to first block on INDEX file
     *  - set return code and position
     */

found_by_membits:
#if 0
    Trace("psf: %3d found in mem; block# %d + %d\n", xid, first, want);
#endif
    pos = (off_t)npos * ESIZE;

    c = 0;
try_pos3:
    c++;
#if 0
    dst = lseek(sfd, pos, SEEK_SET);
#endif
    dst = Iseek(sig_pos, sfd, pos);
    if(dst<(off_t)0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_pos3;
        }
Error("psf: cannot seek %lld on index file fd# %d :%d (%d)",
    (long long)pos, sfd, c, errno);

        ret = XSK_BODY_EPOS;
        *rpos = npos;
        goto out;
    }
    vpos = dst;
    POSVERIFY(sfd, vpos);

    ret = 0;
    *rpos = npos;

    goto found;

found:
 if(DFC) {
    Trace(
"psf: 5 dir# %d, fd# %d, want %d, hash %d:%d:%d:%d range [%d,%d) ; return %d found [%d..%d)\n",
    sig_pos, sfd, want, e->h1, e->h2, e->h3, e->h4,
    ofile_offset(e), (ofile_offset(e)+OFILE_NBLK_SEEK)%OFILE_NBLK_FILE,
    ret, *rpos, (*rpos+want)%OFILE_NBLK_FILE);
 }
    goto out;

not_found:
    dir_lack[fpos]++;
    *rpos = npos;
    ret = XSK_INDEX_NOTFOUND;

 if(DFC) {
    Trace(
"psf: 6 dir# %d, fd# %d, want %d, hash %d:%d:%d:%d range [%d,%d) ; return %d not-found [%d..%d)\n",
    sig_pos, sfd, want, e->h1, e->h2, e->h3, e->h4,
    ofile_offset(e), (ofile_offset(e)+OFILE_NBLK_SEEK)%OFILE_NBLK_FILE,
    ret, *rpos, (*rpos+want)%OFILE_NBLK_FILE);
 }

    goto out;


out:

#if 0
#endif
#ifdef TRACE_INDEX
    pos = (off_t)*rpos * ESIZE;
Trace(
"psf: 7 [%03d:%03d:%03d:%04d] sfd# %d, try %d, match %d, pos %d %12lld <%12lld>, return %d\n",
    e->h1, e->h2, e->h3, e->h4, sfd, OFILE_NBLK_SEEK - try, match,
    *rpos, (long long) pos, (long long) vpos, ret);
#endif

    return ret;
}


int
_pso_match(int sig_pos, int start, int value, int verbose)
{
    int ret;
    int i;
    int xv;

    ret = -1;

#if 0
 if(verbose) {
    Trace(" :pso dir# %d, value %d\n", sig_pos, value);
 }
#endif

    for(i=start;i<OFILE_NBLK_FILE;i++) {
        xv = DMAP_V(sig_pos, i);
#if 0
        if(verbose)
            Trace(" omap: dir# %d, %4d: %10d\n", sig_pos, i, xv);
#endif
        if(xv==value) {
#if 0
            if(verbose)
                Trace("   found value %d in dir# %d, %d\n", value, sig_pos, i);
#endif
            ret = i;
            break;
        }
    }

 if(verbose) {
#if 0
    Trace("  :O dir# %d, value %d found in pos %d ; use %d unuse %d %.2f%%\n",
        sig_pos, value, ret, u, n, 100.0*u/(u+n));
#endif
    Trace("  :O dir# %d, value %d found in pos %d\n",
        sig_pos, value, ret);
 }

    return ret;
}


/*
 * return:
 *      found       0
 *      not found   >0
 *      error       <0
 */

int
bfd_seekobj(int *rfd, bucket_t *bk, skcache_entry_t *e,
    int *rpos, skcache_entry_t *re)
{
    int ret;
    int sfd;
    int c;
    off_t pos, dst;
    int fpos;
    int npos;
    int nr;
    skcache_entry_t *tmp_entry;
    int match = 0;
    int try;
    off_t vpos;
    time_t diff;
    skcache_entry_t chunk[HARD_NINDEX_DREAD];
    int wid, nwid;
    int i;
    int use_map=0;

    int sig_pos;
    int cret;
    int chk;

/*
    int skip;
*/


    *rpos = (off_t)-1;
    ret = 0;
    sfd = *rfd;
    fpos = FDPOOL_POS_E(e);
    sig_pos = DSIG_POS(e->h1, e->h2, e->h3);

#if 0
    Trace("bfd_seekobj: [%03d:%03d:%03d:%04d] len %ld/%ld\n",
        e->h1, e->h2, e->h3, e->h4, (long)e->totallength, (long)OFILE_BLKSIZE);
#endif

 if(DFC) {
Trace(
"pso: dir# %d, sfd# %d, %d:%d:%d:%d from pos %d ; count %d\n",
    sig_pos, sfd, e->h1, e->h2, e->h3, e->h4, npos, dir_count[sig_pos]);
 }

#if 1
    if(dir_count[sig_pos]<0) {
        /* nothing */
        /* following to disk seek */
        npos = ofile_offset(e);
    }
    else
    if(dir_count[sig_pos]==0) {
        ret = XSK_INDEX_NOTFOUND;
        *rpos = -1;
        goto out;
    }
    else {
        use_map = 1;

        npos = _pso_match(sig_pos, 0, OMAP_FUNC(e), DFC);
        if(npos >=0) {
            /* NOTE: found such object.  however it is only compared h4.
             * you should compare other attributes.
             * and maybe exist more matched objects
             */
        }
        else {
#if 0
            ret = XSK_INDEX_NOTFOUND;
            *rpos = -1;
            goto out;
#endif
            goto not_found;
        }
    }
#endif


    try = OFILE_NBLK_SEEK;
    match = 0;
    do {
#if 0
        if(use_map) {
            npos = _pso_match(sig_pos, npos+1, OMAP_FUNC(e), DFC);
 if(DFC) {
            Trace("  :O seek dir# %d, got %d\n", sig_pos, npos);
 }
            if(npos<0) {
                goto not_found;
            }
        }
#endif
        pos = (off_t)npos * ESIZE;
        nwid = MIN(npos+nindex_dread, OFILE_NBLK_FILE) -  npos;
        wid = nwid * ESIZE;
        Iseek(sig_pos, sfd, pos);


        /* block read */
        c = 0;
try_read:
        c++;
 if(DFC) {
Trace("  : dir# %d, %c pos %4d %4d, try %3d: wid %d\n",
            sig_pos, (use_map ? 'm':'-'), npos, npos%OFILE_NBLK_FILE, try, nwid);
 }

        nr = Iread(sig_pos, sfd, chunk, wid);
        if(nr<0) {
            if(errno==EINTR && c < DISK_RETRY) {
                goto try_read;
            }
            Error("bfd_seekobj: cannot read fd# %d pos %lld: %d (%d)",
                sfd, (long long)pos, c, errno);
        }
        vpos += nr;
        nwid = nr/ESIZE;

        for(i=0;i<nwid && try;i++) {
            pos = (off_t)npos * ESIZE;
            tmp_entry = &chunk[i];

#ifdef TRACE_INDEX
#endif
 if(DFC) {
            diff = tmp_entry->issue_time - min_date;

Trace("  :o dir# %d, %c pos %4d %4d, try %3d: %d: %02x %02x, oid# %d, hash %d:%d:%d:%d diff %d\n",
            sig_pos, (use_map ? 'm':'-'),
            npos, npos%OFILE_NBLK_FILE, try,
            DBIT_ISSET(sig_pos, npos),
            (unsigned char)tmp_entry->magic[0],
            (unsigned char)tmp_entry->magic[1],
            tmp_entry->oid,
            tmp_entry->h1, tmp_entry->h2, tmp_entry->h3, tmp_entry->h4,
            diff
            );
 }
            if(tmp_entry->magic[0]==MAGIC_0 && tmp_entry->magic[1]==MAGIC_1) {
                if(tmp_entry->h3==e->h3 && tmp_entry->h4==e->h4) {
                    cret = strcmp(tmp_entry->name,e->name);
 if(DFC) {
                    if(cret!=0) {
                        Trace("  :mem  '%s'\n  :disk '%s' => %d\n",
                            e->name, tmp_entry->name, cret);
                    }
 }
                    if(cret==0) {
                        *re = *tmp_entry;   /* copy */
                        match = 1;
                        break;
                    }
                }
            }

next_blk:
            if(use_map) {
                npos = _pso_match(sig_pos, npos+1, OMAP_FUNC(e), DFC);
                if(npos<0) {
                    goto not_found;
                }
            }
            else {
                npos = (npos+1)% OFILE_NBLK_FILE;
            }

            try--;
#if 0
            if(npos==0) {
                break;
            }
#endif
        }
        if(match || try<=0) {
            break;
        }
    
        if(npos==0) {
#ifdef TRACE_INDEX
            Trace("  : rewind sfd# %d,\n", sfd);
#endif

            c = 0;
try_rewind:
            c++;
#if 0
            dst = lseek(sfd, 0, SEEK_SET);
#endif
            dst = Iseek(sig_pos, sfd, 0);
            if(dst<(off_t)0) {
                if(errno==EINTR && c<DISK_RETRY) {
                    goto try_rewind;
                }
    Error("bfd_seekobj: cannot rewind %lld on index file fd# %d :%d (%d)",
        (long long)pos, sfd, c, errno);
                break;
            }
            vpos = dst;
        }
    } while (match==0 && try>0);

    if(match) {
found:
#if 0
Trace("bfd_seekobj: sfd# %d, found, [%d:%d:%d:%d] npos %d, pos %lld, vpos %ld\n",
    sfd, e->h1, e->h2, e->h3, e->h4,
    npos, (long long)pos, (long long) vpos);
#endif

        c = 0;
try_back:
        c++;
#if 0
        dst = lseek(sfd, - (vpos-pos), SEEK_CUR);
#endif
        dst = Iseek(sig_pos, sfd, pos);
        if(dst<(off_t)0) {
            if(errno==EINTR && c<DISK_RETRY) {
                goto try_back;
            }
    Error("bfd_seekobj: cannot back %lld on index file fd# %d :%d (%d)",
        (long long)pos, sfd, c, errno);

            ret = XSK_INDEX_EROLL;
            *rpos = npos;
            goto out;
        }
        vpos = dst;

        ret = 0;
        *rpos = npos;

#if 0
#endif
 if(DFC) {
Trace(
"pso: dir# %d sfd# %d, hash %d:%d:%d:%d range [%d..%d) ; return %d found %d\n",
        FDPOOL_POS_E(e), sfd, e->h1, e->h2, e->h3, e->h4,
        ofile_offset(e), (ofile_offset(e)+OFILE_NBLK_SEEK)%OFILE_NBLK_FILE, ret, *rpos);
 }
    }
    else {
not_found:
        ret = XSK_INDEX_NOTFOUND;
        *rpos = npos;
#if 0
#endif
 if(DFC) {
Trace(
"pso: dir# %d sfd# %d, hash %d:%d:%d:%d range [%d..%d) ; return %d not-found %d\n",
        FDPOOL_POS_E(e), sfd, e->h1, e->h2, e->h3, e->h4,
        ofile_offset(e), (ofile_offset(e)+OFILE_NBLK_SEEK)%OFILE_NBLK_FILE, ret, *rpos);
 }
    }

    goto out;


out:

#ifdef TRACE_INDEX
    pos = (off_t)*rpos * ESIZE;
Trace("bfd_seekobj: [%03d:%03d:%03d:%04d] *rfd# %d, try %d, match %d, pos %d %12lld <%12lld>, return %d\n",
    e->h1, e->h2, e->h3, e->h4, *rfd, OFILE_NBLK_SEEK - try, match,
    *rpos, (long long) pos, (long long) vpos, ret);
#endif

    return ret;
}




/* 
    return:
        1   fail
        0   success
 */
int
write_attach_fd(int *rifd, int *rbfd, int *rnpos,
    bucket_t *bk, skcache_entry_t *e)
{
    int ret;
    int chk;
    int fpos;
    int npos;
    off_t bipos, bppos; 
    int ifd, bfd;
    int c;

    ret = 0;
    fpos = FDPOOL_POS_E(e);

    *rifd = *rbfd = -7;
    *rnpos = -1;

    ifd = bfd = -3;


    chk = ifd_open(&ifd, bk, e);
    if(chk) {
        Trace("write_attach_fd: fail open ifd# %d, chk %d\n", ifd, chk);
        *rifd = -1;
        ret = chk;

        goto out;
    }

    *rifd = ifd;

    chk = bfd_seekfree(ifd, bk, e, &npos);
    if(chk) {
#if 0
        Trace("write_attach_fd: not found or error <%d>\n", chk);
#endif
        *rnpos = -1;
        ret = 1;

        goto out;
    }
    *rnpos = npos;
    bipos = npos*ESIZE;
    bppos = npos*OFILE_BLKSIZE;


    chk = bfd_open(&bfd, bk, e);
    if(chk) {
        Error("write_attach_fd: fail open bfd# %d, chk %d", bfd, chk);
        *rbfd = -1;
        ret = 1;

        goto out;
    }

    *rbfd = bfd;


    c = 0;
try_lseek:
    c++;
    chk = Pseek(fpos, bfd, bppos);
    if(chk<0) {
Error("write_attach_fd: fail lseek at npos %ld %lldB in bfd# %d, chk %d (%d)",
(long)npos, (long long)bppos, bfd, chk, errno);
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_lseek;
        }
        *rnpos = -1;
        ret = 1;
    }
    else {
        ret = 0;
    }


out:
#if 0
Trace("write_attach_fd: done *rifd# %d, *rbfd# %d, pos %ld %lldB, ret %d\n",
        *rifd, *rbfd, (long)*rnpos, (long long)bppos, ret);
#endif

    return ret;
}



/* 
    return:
        1   fail
        0   success
 */

int
read_attach_fd(int *rifd, int *rbfd, int *rnpos,
    bucket_t *bk, skcache_entry_t *e, skcache_entry_t *re)
{
    int ret;
    int chk;
    off_t vpos;
    int fpos;
    int npos;
    int ifd, bfd;
    off_t bpos;
    int c;

    ret = 0;
    fpos = FDPOOL_POS_E(e);


    *rifd = *rbfd = -25;
    *rnpos = -1;
    
    ifd = bfd = -57;

    chk = ifd_open(&ifd, bk, e);
    if(chk) {
        Error("read_attach_fd: fail open ifd# %d, chk %d", ifd, chk);

        *rifd = -1;
        ret = chk;

        goto out;
    }

    *rifd = ifd;

    chk = bfd_seekobj(&ifd, bk, e, &npos, re);
    if(chk) {
        if(chk>0) {
            /*
             * MISS - if 'chk' is setted, it is not error.
             */
#if 0
            Trace("read_attach_fd: MISS, not found the object <%d>\n", chk);
#endif
            *rnpos = -1;
            ret = chk;
        }
        else {
            Error("read_attach_fd: fail bfd_seekobj() <%d>", chk);
            *rnpos = -1;
            ret = -1;
        }

        goto out;
    }

    *rnpos = npos;
    bpos = npos*OFILE_BLKSIZE;

    chk = bfd_open(&bfd, bk, e);
    if(chk) {
Trace("read_attach_fd: fail open bfd# %d, chk %d\n", bfd, chk);

        *rbfd = -1;
        ret = chk;

        goto out;
    }
        
#if 0
    {
        off_t v, v2;
        v = lseek(ifd, 0, SEEK_CUR);
Trace("read_attach_fd: v %d %lldB on ifd# %d %d\n",
    npos, (long long)v, ifd, *rifd);

        dump_ifile(ifd, ofile_offset(e));

        v2 = lseek(ifd, v, SEEK_SET);
    }
#endif


    *rbfd = bfd;

    c = 0;
try_lseek:
    c++;
#if 0
    vpos = lseek(bfd, bpos, SEEK_SET);
#endif
    vpos = Pseek(fpos, bfd, bpos);
    if(vpos<0) {
Error("read_attach_fd: fail lseek at %ld %lld in bfd# %d, vpos %d (%d)",
    (long)npos, (long long)bpos, bfd, vpos, errno);
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_lseek;
        }
        *rnpos = -1;
        ret = 1;
    }
    else {
#if 0
Trace("read_attach_fd: done lseek at pos %ld %lldB in bfd# %d, vpos %lldB\n",
    (long)npos, (long long)bpos, bfd, (long long)vpos);
#endif
        ret = 0;
    }


out:
#if 0
Trace("read_attach_fd: done *rifd# %d, *rbfd# %d, pos %ld %lldB, ret %d\n",
        *rifd, *rbfd, (long)*rnpos, (long long)bpos, ret);
#endif

    return ret;
}




int
detach_fd(int *rfd)
{
#ifndef OBJECT_PACKING
    /*
    bfd_close(rfd);
    */
#else
    (void)1;
#endif
}


/*
 * return
 *      0   success
 *      >0  fail
 *
 */

int
skcache_entry_swapin(bucket_t *bk, skcache_entry_t *e)
{
    int ifd, bfd;
    int npos;
    skcache_entry_t tmp_entry;
    skcache_frag_t *cur, *pre, *dt;
    int nr;
    off_t nr_sum;
    int chk;
    int i;
    int ret;
    int nf;
    struct timeval begin, end;
    int c;
    off_t fsize;
    off_t tmpfsize;
    int erase;
    int sigpos;
    int lf;
    int er;
    int rest;

    int want;
    struct iovec iov[HARD_NFRAG_DREAD];
    skcache_frag_t *chunk[HARD_NFRAG_DREAD];
    off_t iovlen;
    int iovnum; 
    off_t tsum;

    int takes;
    int fsnum;


    gettimeofday(&begin, NULL);
    ret = 0;

    FCHK();

#if 0
    Trace("swapin: b bk# %d, entry %p, oid# %d, d_recorded %d, m_recorded %p, count %d, len %d\n",
        bk->bid, e, e->oid, e->d_recorded, e->m_recorded, e->count, e->totallength);
#endif

    assert(sk_entity_check(bk, __FILE__, __LINE__));

    if(e->m_recorded!=NULL) {
        Error("swapin: not recorded in memory bk# %d, e %p", bk->bid, e);
#if 0
        ret = XSK_EIGNORE;
#endif
        ret = XSK_SWAP_EMPTY;
        goto out;
    }

    sigpos = DSIG_POS(e->h1, e->h2, e->h3);
    fsnum = bk->fs;

    MUTEX_LOCK(&stat_lock);
    swapin_count++;
    fs_swapin_count[fsnum]++;
    MUTEX_UNLOCK(&stat_lock);

    MUTEX_LOCK(&dir_mutex);
    lf = dir_lock[sigpos];
    MUTEX_UNLOCK(&dir_mutex);
    if(lf>0) {
/*
        Trace("swapin: conflict directory bk# %d, dir# %d,.\n", bk, sigpos);
*/
        MUTEX_LOCK(&stat_lock);
        dir_conflict++;
        MUTEX_UNLOCK(&stat_lock);

        ret = XSK_SWAP_ECONF;
    }
    if(ret) {
        goto out;
    }

    chk = read_attach_fd(&ifd, &bfd, &npos, bk, e, &tmp_entry);
    if(chk>0) {
#if 0
        Trace("swapin: not found object (chk=%d)\n", chk);
#endif
        ret = XSK_SWAP_MISS;
        goto out;
    }
    if(chk<0 || ifd<0 || bfd<0 || ifd==bfd || npos<0) {
        Error("swapin: error chk %d, ifd# %d, bfd# %d, pos %ld",
            chk, ifd, bfd, (long)npos);
        ret = XSK_SWAP_EATTACH;
        goto out;
    }



    /***
     *** check contents
     ***/

    /*
     * this routing MUST check magic.
     * because disk maybe stores broken or older version cache file. 
     */
    if(tmp_entry.magic[0]!=MAGIC_0
        || tmp_entry.magic[1]!=MAGIC_1) {
#ifdef CHECK_MAGIC
        Error("swapin: ignore magic %02x %02x / %02x %02x '%s'",
            tmp_entry.magic[0], tmp_entry.magic[1],
            MAGIC_0, MAGIC_1, tmp_entry.name);
#endif
        ret = XSK_EMAGIC;
        goto fail_read_cntl;
    }

    tmp_entry.m_recorded = NULL;
#if 0
    Trace("swapin: read ifd# %d, npos %d, oid# %d, [%d:%d:%d:%d] %s\n",
        ifd, npos, 
        tmp_entry.oid, tmp_entry.h1, tmp_entry.h2,
        tmp_entry.h3, tmp_entry.h4, tmp_entry.name);
    Trace("  target is [%d] '%s'\n", e->h4, e->name);
#endif

    if(tmp_entry.oid<=0) {
#if 0
        Trace("swapin: reject negative oid in %p, oid# %d,\n",
            &tmp_entry, tmp_entry.oid);
#endif
        ret = XSK_EOID;
        goto fail_read_cntl;
    }

    if(tmp_entry.h4<0) {
#if 0
        Trace("swapin: reject negative h4 in %p, h4 %d,\n",
            &tmp_entry, tmp_entry.h4);
#endif
        ret = XSK_EH4;
        goto fail_read_cntl;
    }

    if(strcmp(tmp_entry.name,e->name)!=0) {
#if 0
        Trace("swapin: read oid# %d, hash %d,%d,%d,%d %s\n",
            tmp_entry.oid, tmp_entry.h1, tmp_entry.h2,
            tmp_entry.h3, tmp_entry.h4, tmp_entry.name);
        Trace("but target is '%s'\n", e->name);
#endif

#if 0
        Trace("    oid# %d, hash %d,%d,%d,%d\n",
            tmp_entry.oid, tmp_entry.h1, tmp_entry.h2,
            tmp_entry.h3, tmp_entry.h4);
        Trace("    saved %s\n    wont  %s\n", tmp_entry.name, e->name);
        {
            int yh1, yh2, yh3, yh4;
            calc_hash(tmp_entry.name,   &yh1, &yh2, &yh3, &yh4);
            calc_hash(e->name,          &yh1, &yh2, &yh3, &yh4);
        }
#endif

        ret = XSK_EHASH;
        erase = 1;
        goto fail_read_cntl;
    }

    if(tmp_entry.issue_time<0) {
#if 1
        Error("swapin: ignore issue_time");
#endif
        ret = XSK_EISSUE;
        goto fail_read_cntl;
    }

    if(tmp_entry.issue_time<min_date) {
#if 1
        Error("swapin: very old object");
#endif
        ret = XSK_EOLD;
        goto fail_read_cntl;
    }


    /***
     *** correct entry.
     ***/
    {
        int old_oid;
        void *old_owner;

        old_oid = e->oid;
        old_owner = (void*)e->owner;

        *e = tmp_entry;

        (void*)e->owner = old_owner;
        e->d_recorded = e->oid;
    }


    
skip_read_cntl:
    (void)1;


#if 0
    Trace("oid# %d, totallength %lld\n",
        e->oid, (long long)e->totallength);
#endif


    want = NBCEIL(e->totallength, OFILE_BLKSIZE);

    for(i=0;i<HARD_NFRAG_DREAD;i++) {
        chunk[i] = NULL;
    }


    pre = NULL;

    nr_sum = (off_t)0L;
    nf = 0;                 /* the number of frag */
    do {
        rest = e->totallength - nr_sum;
#if 0
        Trace("oid# %d, rest %lld, totallength %lld, nr_sum %lld: dread %d\n",
            e->oid,
            (long long)rest, (long long)e->totallength, (long long)nr_sum,
            nfrag_dread);
#endif
        iovnum = 0;
        iovlen = 0L;
        for(i=0;i<nfrag_dread && iovlen < rest ;i++) {
            cur = skcache_pull_new_frag(bk);
            if(cur==NULL) {
                ret = XSK_NOFRAG;
                goto fail_read;
            }
            bk->own_frag++;

            chunk[i] = cur;
            iov[i].iov_base = cur->ent;
            if(iovlen + cur->size > rest) {
                iov[i].iov_len  = rest - iovlen;
                iovlen += iov[i].iov_len;
            }
            else {
                iov[i].iov_len  = cur->size;
                iovlen += cur->size;
            }
            iovnum++;
        }
        cur->next = NULL;

#if 0
    Trace("oid# %d, iov num %d len %lld\n",
        e->oid, iovnum, (long long)iovlen);
#endif

        c = 0;
try_read:
        nr = Preadv(sigpos, bfd, iov, iovnum);

#if 0
        Trace("<<< %3d %p: read %5d/%ld %7ld? oid# %d, %lld\n",
            nf, cur, nr, (long)iovlen,
            (nr>0) ? (long)nr_sum+nr : (long)nr_sum,
            e->oid, (long long)e->totallength);
#endif


        if(nr<0) {
            if(errno==EINTR && c<DISK_RETRY) {
                goto try_read;
            }

            Error("swapin: data, fail read() return %d %d (%d)",
                nr, c, errno);

#if 0
            cur->len = 0;
            skcache_push_idle_frag(bk, cur);
            bk->own_frag--;
            cur = NULL;
#endif

            ret = XSK_SWAP_EREAD;
            goto fail_read;
        }
#if 0
        if(nr==0) {
#if 0
            Trace("  oid# %d, reach end\n", e->oid);
#endif
            cur->len = 0;

            /*
             * When only read first frag, keep the frag.
             * otherwise, remove these frags to save memory.
             */
            if(nf>0) {
                skcache_push_idle_frag(bk, cur);
                bk->own_frag--;
                cur = NULL;
            }
            
            break;
        }
#endif

        if(nr!=iovlen) {
            Error("swapin: oid# %d, read act/spc %lld/%lld\n",
                e->oid, nr, iovlen);
        }

        /* append read chunk to past read */
        tsum = 0;
        for(i=0;i<iovnum;i++) {
            cur = chunk[i];
            cur->len = iov[i].iov_len;

            if(pre==NULL) {
                if(e->m_recorded) {
Error("oid# %d, i over write %p upon %p", e->oid, cur, e->m_recorded);
                }
                e->m_recorded = cur;
            }
            else {
                pre->next = cur;
            }
            pre = cur;
            cur->next = NULL;

            nr_sum += iov[i].iov_len;
            tsum += iov[i].iov_len;


            if(tsum>=nr) {
                break;
            }
        }
        if(nr==0) {
            break;
        }

        nf++;
    } while( nr_sum < e->totallength );
 
    detach_fd(&bfd);

#if 0
    Trace("  swapin 1 read %lld/%lld %lld oid# %d, '%s'\n",
        (long long)nr_sum, (long long)e->totallength,
        (long long)tmp_entry.totallength,
        e->oid, e->name);
#endif

    /*
     * fix length and empty frags in chunk
     */

#if 0
    Trace("  swapin 2 read %lld/%lld %lld oid# %d, '%s'\n",
        (long long)nr_sum, (long long)e->totallength,
        (long long)tmp_entry.totallength,
        e->oid, e->name);
#endif

    if(nr_sum != e->totallength) {
        Error("swapin: oid# %d, total read act/spc %lld/%lld '%s'",
            e->oid,
            (long long)nr_sum, (long long)e->totallength,
            e->name);
    }

    MUTEX_LOCK(&stat_lock);
    swapin_success++;
    MUTEX_UNLOCK(&stat_lock);

    goto out;

/*
 * recoverely
 */

fail_read_cntl:
#if 0
    Trace("swapin: fail read contorl block reason %d, erase %d, size %ld\n",
        ret, erase, (long)fsize);
#endif
    goto fail;

fail_read:
#if 1
    Trace("swapin: fail to read data. trush just readings, reason %d\n",
        ret);
#endif

#if 0
    Trace("swapin: clear read %d frags, %lld bytes, reason %d\n", 
        nf, (long long)nr_sum, ret);
#endif

    goto fail;

fail:
    detach_fd(&bfd);
   
    if(ret<0) {
        sim_stat_add_value(&swapin_fail_reason, -ret);
    }
    else {
        sim_stat_add_value(&swapin_fail_reason, ret);
    }

    goto out;


out:
#if 0
    Trace("  frag use/all = %d/%d\n", bk->own_frag, bk->frag_all);
#endif

    gettimeofday(&end, NULL);
    takes = KOTETU_SUB_TIME(end, begin);
    sim_stat_add_value(&swapin_time, takes);
    sim_hist_add_value(&swapin_hist, takes);
    sim_hist_add_value(&fs_swapin_hist[fsnum], takes);

#if 0
    Trace("swapin: a bk# %d, entry %p, oid# %d, d_recorded %d, m_recorded %p, count %d, len %d, return %d\n",
        bk->bid, e, e->oid, e->d_recorded, e->m_recorded, e->count, e->totallength, ret);
#endif

    FCHK();

    return ret;
}









/*
 *
 * return
 *      0   success
 *      !0  fail
 *              >0  ignore
 *              <0  temporary, redo again
 */

int
skcache_entry_swapout(bucket_t *bk, skcache_entry_t *e)
{
    int ret;
    int ifd, bfd;
    int npos;
    skcache_frag_t *cur, *pre;
    int nw;
    off_t nw_sum;
    int chk;
    int i, ii;
    struct timeval begin, end;
    struct stat sbuf;
    int sig_pos;
    int c;
    int takes;

    int fsnum;

    struct iovec iov[HARD_NFRAG_DWRITE];
    int     numf;
    off_t   lenf;

    off_t last_size;
    int lf;

    int want;


    gettimeofday(&begin, NULL);

    ret = 0;
    FCHK();

    if(bk==NULL) {
        Error("nil bucket pointer %x", bk);
        ret = XSK_EBUCKET;
        goto fail_start;
    }
    if(e==NULL) {
        Error("nil entry pointer %x", e);
        ret = XSK_EMPTY;
        goto fail_start;
    }

    sig_pos = DSIG_POS(e->h1, e->h2, e->h3);
    fsnum = bk->fs;

#if 0
    Trace("swapout: b bk# %d, entry %p, oid# %d, d_recorded %d, m_recorded %p, count %d, len %ld\n",
        bk->bid, e, e->oid, e->d_recorded, e->m_recorded, e->count, (long)e->totallength);
#endif

    assert(sk_entity_check(bk, __FILE__, __LINE__));

    if(e->m_recorded==NULL) {
      Trace("swapout: not recorded in memory bk# %d, e %p, oid# %d, length %ld\n",
            bk->bid, e, e->oid, (long)e->totallength);
        ret = 0;
        goto out;
    }
    if(e->issue_time==0 || e->oid==0) {
#if 0
        Trace("swapout: temporary entry bk# %d, e %p, oid# %d, length %ld\n",
            bk->bid, e, e->oid, (long)e->totallength);
#endif
        ret = 0;
        goto out;
    }

#ifndef SAVECOUNT
    if(e->d_recorded!=0)
    {
#if 0
    Trace("+# oid# %d, d_recorded# %d, count %3d,\n\thot %d, name %s\n",
        e->oid, e->d_recorded, e->count, e->hot, e->name);
#endif
        ret = 0;
        goto out;
    }
#endif

    if(e->totallength > OFILE_BLKSIZE * OFILE_NBLK_OBJ)
    {
#if 0
        Trace("swapout: oid# %d, file size %ld %ldK is larger than %ld <BLKx%d>\n",
            e->oid,
            (long)e->totallength+ESIZE,
            dK((long)e->totallength+ESIZE),
            (long)OFILE_BLKSIZE*OFILE_NBLK_OBJ, OFILE_NBLK_OBJ);
#endif
        MUTEX_LOCK(&stat_lock);
        count_too_big_object++;
        MUTEX_UNLOCK(&stat_lock);

        ret = 0;
        goto out;
    }

    want = NBCEIL(e->totallength, OFILE_BLKSIZE);

    MUTEX_LOCK(&stat_lock);
    swapout_count++;
    fs_swapout_count[fsnum]++;
    MUTEX_UNLOCK(&stat_lock);

    MUTEX_LOCK(&dir_mutex);
    lf = dir_lock[sig_pos];
    MUTEX_UNLOCK(&dir_mutex);
    if(lf>0) {
/*
        Trace("swapout: conflict directory bk# %d, dir# %d,.\n",
            bk->bid, sig_pos);
*/
        MUTEX_LOCK(&stat_lock);
        dir_conflict++;
        MUTEX_UNLOCK(&stat_lock);

        ret = XSK_SWAP_ECONF;
        goto fail_start;
    }


    chk = write_attach_fd(&ifd, &bfd, &npos, bk, e);
    if(chk || ifd<0 || bfd<0 || ifd==bfd || npos<0) {
/*
*/
#if 0
        Trace(
         "swapout: not found object (%d) or error ifd# %d, bfd# %d, pos %ld\n",
            chk, ifd, bfd, (long)npos);

        Trace("  dir# %d, dir_sz %lld %.2f%% hwater/max=%lld/%lld\n",
            sig_pos, (long long)dir_sz[sig_pos],
            ((double)dir_sz[sig_pos]*100.0/(double)dir_size_max),
            (long long)dir_size_hwater, (long long)dir_size_max
            );
#endif

        ret = XSK_SWAP_NOTFOUND;
        goto fail_open;
    }

#if 0
    Trace("swapout: oid# %d, attach to file ifd# %d, bfd# %d, \n",
        e->oid, ifd, bfd);
#endif

    /***
     *** write control
     ***/

#if 0
    if(want>1) {
        Trace("  *** want %d blocks over %d..%d\n", want, npos, npos+want);
    }
#endif
    for(ii=npos;ii<npos+want;ii++) {
        i = ii%OFILE_NBLK_FILE;

        c = 0;
try_write_cntl1:
        c++;
        nw = Iwrite(sig_pos, ifd, e, ESIZE);
        if(nw<0) {
            if(errno==EINTR && c<DISK_RETRY) {
                goto try_write_cntl1;
            }

            if(errno==ENOSPC) {
                Error("disk full!! at write control in swapout 1");
            }
            else
            if(errno==ENOMEM) {
                Error("kernel said no memory. <%d>", c);

                msleep(1);

                if(c<DISK_RETRY) {
                    goto try_write_cntl1;
                }

            }
            else {
                Error("swo: fail Iwrite() index return %d %d (%d)", nw, c, errno);
            }
            ret = XSK_SWAP_EWRITE;
            goto fail_write_cntl;
        }

    }


    if(e->d_recorded!=0) 
    {
#if 0
    Trace("+# oid# %d, d_recorded# %d, count %3d,\n\thot %d, name %s\n",
        e->oid, e->d_recorded, e->count, e->hot, e->name);
#endif
        goto skip_body_swapout;
    }


    /***
     *** write body
     ***/

    cur = e->m_recorded;
    if(cur==NULL) {
        goto skip_body_swapout;
    }

    nw_sum = 0L;
    i = 0;
    while(cur) {

        /* write frag chunk by writev() */
        numf = 0;
        lenf = (off_t)0L;
        while(cur && numf<nfrag_dwrite) {
#if 0
            Trace(">>> %3d %3d %p: stored %5d/%d %7ld\n",
                i, numf, cur->ent, cur->len, cur->size, (long)nw_sum+cur->len);
#endif
            if(cur->len>0) {
                iov[numf].iov_base  = cur->ent;
                iov[numf].iov_len   = cur->len;
                numf++;

                lenf += (off_t) cur->len;
            }
            cur = cur->next;
        }
#if 1
        for(i=numf;i<HARD_NFRAG_DWRITE;i++) {
            iov[i].iov_base = NULL;
            iov[i].iov_len  = -1;
        }
#endif

        c = 0;
try_write:
        c++;
#if 0
        nw = writev(bfd, iov, numf);
#endif
        nw = Pwritev(sig_pos, bfd, iov, numf);
#if 0
        Trace(">>> %3d write %ld\n",
            i, nw);
#endif

        if(nw<0) {
            if(errno==EINTR && c<DISK_RETRY) {
                goto try_write;
            }
            Error("swo: fail Pwritev() data return %d (%d)", nw, errno);
            if(errno==ENOSPC) {
                Error("disk full!! when swapout data");
            }

            ret = 7;
            goto fail_write;
        }
        else
        if(nw==0) {
            break;
        }
        if(nw>0) {
#if 1
            if(nw!=lenf) {
                Error("swapout: oid# %d, write act/spc %lld/%lld bytes\n", 
                    e->oid, (long long)nw, (long long)lenf);
            }
#endif
            nw_sum += nw;
        }

        i++;
    }
#if 1
    if(nw_sum!=e->totallength) {
        Trace("swapout: oid# %d, total act/spc %lld/ %lld bytes\n",
            e->oid, (long long)nw_sum, (long long)e->totallength);
    }
#endif

    /* XXX */
#if 0
    ftruncate(bfd, (off_t) nw_sum + ESIZE );
#endif

skip_body_swapout:

    detach_fd(&bfd);
    skcache_entry_purge(bk, e);
    
    e->m_recorded = NULL;
    e->d_recorded = e->oid;
#if 1
    e->swapout_candidate = 0;
#endif

#if 0
    Trace("swapout: oid# %d, write %ld, %ld+%ld/%ld\n",
        e->oid, ESIZE + (long)nw_sum,
        ESIZE, (long)nw_sum, (long)e->totallength);
#endif

#if 0
    Trace("swapout: p %p, oid# %d, path '%s', size %ld\n", 
        e, e->oid, name, (long)nw_sum);
#endif

    last_size = 0;

    MUTEX_LOCK(&stat_lock);
    swapout_success++;

    if(last_size<=0) {
        swap_count += 1LL;
        swap_size +=
            (long long)(nw_sum + (off_t)ESIZE);
    }
    else {
        swap_size +=
            (long long)(nw_sum + (off_t)ESIZE - last_size);
    }
#if 0
    Trace("swapout: oid# %d, total swap count %lld size %lld\n",
        e->oid, swap_count, swap_size);
#endif
    MUTEX_UNLOCK(&stat_lock);

#if 0 
    MUTEX_LOCK(&dir_mutex);
#endif

        DSIG_SET(sig_pos, e->h4);

        for(ii=npos;ii<npos+want;ii++) {
            i = ii%OFILE_NBLK_FILE;
            DBIT_SET(sig_pos, i);
            DMAP_SET(sig_pos, i, OMAP_FUNC(e));
        }
    if(dir_count[sig_pos]<0) {
        /* nothing */
    }
    else {
        dir_sz[sig_pos] +=
            (off_t)(nw_sum + (off_t)ESIZE - last_size);
        dir_count[sig_pos]++;
    }

#if 0
    MUTEX_UNLOCK(&dir_mutex);
#endif

#if 0
    dsig_show(&dir_sig[sig_pos*DSIG_ELEM_WIDTH]);
#endif
#if 0
    dbit_show(&dir_bit[sig_pos*DBIT_ELEM_WIDTH]);
#endif

    goto out;

fail_write_cntl:
fail_write:
    detach_fd(&bfd);

    if(ret<0) {
        sim_stat_add_value(&swapout_fail_reason, -ret);
    }
    else {
        sim_stat_add_value(&swapout_fail_reason, ret);
    }

    MUTEX_LOCK(&stat_lock);
    if(last_size<0) {
    }
    else {
        swap_size -= last_size;
    }
    MUTEX_UNLOCK(&stat_lock);

    goto out;

fail_open:
fail_start:

    if(ret<0) {
        sim_stat_add_value(&swapout_fail_reason, -ret);
    }
    else {
        sim_stat_add_value(&swapout_fail_reason, ret);
    }

    goto out;

out:
#if 0
    Trace("  frag use/all = %d/%d\n", bk->own_frag, bk->frag_all);
#endif
    gettimeofday(&end, NULL);
    takes = KOTETU_SUB_TIME(end, begin);
    sim_stat_add_value(&swapout_time, takes);
    sim_hist_add_value(&swapout_hist, takes);
    sim_hist_add_value(&fs_swapout_hist[fsnum], takes);

    assert(sk_entity_check(bk, __FILE__, __LINE__));
#if 0
    Trace("swapout: a bk# %d, entry %p, oid# %d, d_recorded %d, m_recorded %p, count %d, len %ld, return %d\n",
        bk->bid, e, e->oid, e->d_recorded, e->m_recorded, e->count, (long)e->totallength, ret);
#endif

    FCHK();

    return ret;
}





#ifdef NO_VS
#ifdef VS
#undef VS
#endif
#endif




int
rm_fr(char *base)
{
    char tmp[L_PATH];
    dirent_t    *dp;
    DIR         *dirp;
#ifdef PC_NAME_MAX
    dirent_t ent[PC_NAME_MAX+1];
#else
#ifdef _PC_NAME_MAX
    dirent_t ent[_PC_NAME_MAX+1];
#else
    dirent_t ent[128];
#endif
#endif
    struct stat sbuf;
    int chk;

    if(base==NULL)
        return -1;

#if 0
Trace("rm_fr: called '%s'\n", base);
#endif

    chk = rmdir(base);
    if(chk) {
        if(errno==ENOTEMPTY) {

            dirp = opendir(base);
            if(dirp) {
                while(1) {
                    chk = readdir_r(dirp, ent, &dp);
                    if(dp==NULL) {
                        break;
                    }

                    if(dp->d_name[0]=='.' && dp->d_name[1]=='\0') {
                        continue;
                    }
                    if(dp->d_name[0]=='.'
                        && dp->d_name[1]=='.' && dp->d_name[2]=='\0') {
                        continue;
                    }
                    /*
                    strcpy(tmp, base);
                    strcat(tmp, "/");
                    strcat(tmp, dp->d_name);
                    */
                    pathncat(tmp, L_PATH, base, dp->d_name);


                    chk = stat(tmp, &sbuf);
                    if(chk) {
                        Error("rm_fr: fail stat() '%s' (%d)",
                            tmp, errno);
                        continue;
                    }

                    if(S_ISREG(sbuf.st_mode)) {
                        unlink(tmp);
                    }
                    if(S_ISDIR(sbuf.st_mode)) {
                        rm_fr(tmp);
                    }
                }
                closedir(dirp);
            }

            chk = rmdir(base);

            if(chk) {
                Error("rm_fr: fail rmdir() '%s' after sweep (%d)",
                    base, errno);
            }

        }
    }

    return 0;
}





extern pthread_mutex_t stat_lock;
extern volatile int compaction_count;

typedef struct {
    int no;
    int type;   /* 0 single, 1 packed */
    int pos;    /* position */
    int act;
    time_t date;
    off_t sz;
} tup;

int
comp_tup(const void* xp, const void* yp)
{
    tup *x = (tup*)xp;
    tup *y = (tup*)yp;

    if(x->act == y->act) {
        if(x->date == y->date) {
            return x->sz - y->sz;           /* sz smaller */
        }
        else {
            return y->date - x->date;       /* date grater */
        }
    }
    else {
        return x->act - y->act;             /* act smaller */
    }
}

int
comp_tup_order(const void* xp, const void* yp)
{
    tup *x = (tup*)xp;
    tup *y = (tup*)yp;

    if(x->type == y->type) {
        return x->pos - y->pos;             /* y-x smaller */
    }
    else {
        return y->type- x->type;            /* y-x greater */
    }
}

int
comp_tup_time(void*xp,void*yp)
{
    tup *x = (tup*)xp;
    tup *y = (tup*)yp;

    return x->date - y->date;
}

int
show_tup(char *msg, int num, tup *base)
{
    int i;
    tup *cur;
    off_t lsum, osum;

    Trace("show_tup: %s\n", msg);

    lsum = osum = (off_t) 0L;
    cur = base;
    for(i=0;i<num;i++) {
        if(cur->act==ACTION_NOP || cur->act==ACTION_SKIP) {
        lsum += cur->sz;
        Trace(" %12d: %c %4d : %d %10d %12lld %12lld\n",
            cur->no, cur->type ? 'P' : 's', cur->pos,
            cur->act, cur->date, (long long)cur->sz, (long long)lsum);
        }
        else {
        osum += cur->sz;
        Trace(" %12d: %c %4d : %d %10d %12lld\n",
            cur->no, cur->type ? 'P' : 's', cur->pos,
            cur->act, cur->date, (long long)cur->sz);
        }
        cur++;
    }
}


int 
decision_disk_action(skcache_entry_t *en, time_t reftime, int *xaction)
{
    int xh1, xh2, xh3, xh4;
    int igh;

    *xaction = ACTION_SKIP;

#if 0
Trace("check_target: en %p, MAGIC %x %x, oid# %d, *xaction %d\n",
    en, (unsigned char)en->magic[0], (unsigned char)en->magic[1],
    en->oid, *xaction);
#endif

    if(en->magic[0]!=MAGIC_0 || en->magic[1]!=MAGIC_1) {
        *xaction = ACTION_PURGE;
#ifdef CHECK_MAGIC
Error("check_target: oid# %d, wrong magic %02x %02x / %02x %02x",
        en->oid, 
        en->magic[0], en->magic[1], MAGIC_0, MAGIC_1);
#endif
        en->issue_time = -1;
        goto out;
    }

    if(en->oid == 0) {
#if 0
Trace("oid# %d, *xaction %d -> 0\n", en->oid, *xaction);
#endif
        *xaction = ACTION_PURGE;
        goto out;
    }

    calc_hash(en->name, &xh1, &xh2, &xh3, &xh4);

    if(en->h1!=xh1 || en->h2!=xh2 || en->h3!=xh3 || en->h4!=xh4) {
        *xaction = ACTION_PURGE;
        igh++;
#if 0
Trace("\
check_target: ignore hash, file '%s'\n\
    stored %03d:%03d:%03d:%03d %s\n\
    solved %03d:%03d:%03d:%03d same\n",
    name,
    en->h1, en->h2, en->h3, en->h4, en->name,
    xh1, xh2, xh3, xh4);
#endif
#if 0
Trace("check_target: ignore hash, stored %03d:%03d:%03d:%03d, we guess %03d:%03d:%03d:%03d\n",
        en->h1, en->h2, en->h3, en->h4, xh1, xh2, xh3, xh4);
#endif
        en->issue_time = -1;
        goto out;
    }

/*
    idiff = reftime - en->issue_time;
    ldiff = reftime - en->lmd;
    ediff = reftime - en->expire;
*/


    if(en->issue_time<=0) {
        *xaction = ACTION_PURGE;
    }
    else {
        /* black */
        if(en->issue_time<min_date) {
            *xaction = ACTION_PURGE;
        }
        else
        if(en->issue_time>reftime) {
            *xaction = ACTION_PURGE;
        }
        else
        if(en->expire>0 && reftime >= en->expire ) {
            *xaction = ACTION_PURGE;
        }

        /* gray */
    }

out:
    return *xaction;
}


int check_target(skcache_entry_t *en, time_t reftime, int *xaction)
{
    return decision_disk_action(en, reftime, xaction);
}


/*
 * convert num-num -> num
 */
int name2num(char *nm)
{
    int r1, r2, r3, r4;
    int ret;
    register char *p, *q;
    register int i;
    char tmp[L_PATH];
    
    p = nm;

    q = tmp;
    i = 0;
    while(*p && (*p>='0' && *p<='9') && i < L_PATH) {
        *q++ = *p++;
        i++;
    }
    p++;
    *q = '\0';
    r3 = atoi(tmp);
    
    q = tmp;
    i = 0;
    while(*p && (*p>='0' && *p<='9') && i < L_PATH) {
        *q++ = *p++;
        i++;
    }
    p++;
    *q = '\0';
    r4 = atoi(tmp);

    ret = r3*radix_h4 + r4;

#if 0
    Trace("name2num '%s' -> %d\n", nm, ret);
#endif

    return ret;
}


#include "chkdir.c"


volatile int dsweeper_seq   = 0;
volatile int dsweeper_count = 0;
volatile int dsweeper_phase = -1;
volatile int dsweeper_pos   = -1;
volatile int dsweeper_full_mode = 0;

/* how many tics are required to work sweeper */
#ifndef SWEEP_CYCLE
#define SWEEP_CYCLE (1)
#endif
#ifndef SWEEP_BOOT_CYCLE
#define SWEEP_BOOT_CYCLE (1)
#endif

/* how many direcotry are sweeped per chance */
#ifndef SWEEP_NTRY
#define SWEEP_NTRY  (1)
#endif


volatile int dsweep_boot_cycle = SWEEP_BOOT_CYCLE;
volatile int dsweep_cycle = SWEEP_CYCLE;



/* for two level */
int
skcache_dir_sweeper(void)
{
    int f;
    int d1, d2, d3;
    int v;
    int chk;
    int p;
    int nd1;
    int pos;

    MUTEX_LOCK(&stat_lock);
    if(dsweeper_phase<0) {
        dsweeper_phase = 0;
        dsweeper_pos = rand() % dir_num;
    }
    p = dsweeper_phase;
    f = dsweeper_seq;
    dsweeper_seq =  (dsweeper_seq+1) % 1000000;

#if 0
    if(dsweeper_count%dir_num==0) {
        Trace("first dir\n");
    }
#endif
    dsweeper_count++;

    v = dsweeper_pos;

    if(dsweeper_pos>=dir_num-1) {
        dsweeper_phase++;
        dsweeper_pos = 0;
    }
    else {
        dsweeper_pos++;
    }
    MUTEX_UNLOCK(&stat_lock);

    if(f<0) {
        f = v;
    }

    d1  = v%radix_h1;
    d2  = v/radix_h1;
    d3  = 0;

    pos = DSIG_POS(d1, d2, d3);

#if 0
    Trace("  sweeper th %08x %6d %6d -> dir# %d, %03d %03d %03d\n",
        pthread_self(), p, v, pos, d1, d2, d3);
#endif
    MUTEX_LOCK(&dir_mutex);
    if(dir_lock[pos]>0) {
        pos = -1;
    }
    else {
        dir_lock[pos] = pthread_self();
    }
    MUTEX_UNLOCK(&dir_mutex);

    if(pos>=0) {

        chk = skcache_check_dir(d1, d2, d3, 1);

        MUTEX_LOCK(&dir_mutex);
        dir_lock[pos] = 0;
        MUTEX_UNLOCK(&dir_mutex);
    }

done:

    return 0;
}


static
void
skcache_preload_files()
{
    int i, j;
    skcache_entry_t dummy_ent;
    int dummy_ifd, dummy_bfd;
    int chk;

    Trace("skcache_preload_files: preload %d files (%d x %d x 2)\n",
        radix_h1 * radix_h2 * 2, radix_h1, radix_h2);
    for(i=0;i<radix_h1;i++) {
        for(j=0;j<radix_h2;j++) {
            dummy_ent.h1 = i;
            dummy_ent.h2 = j;

            chk = ifd_open(&dummy_ifd, NULL, &dummy_ent);
            chk = bfd_open(&dummy_bfd, NULL, &dummy_ent);
#if 0
            Trace(" (i,j) %4d %4d -> (ifd, bfd) %5d %5d\n",
                i, j, dummy_ifd, dummy_bfd);
#endif
        }
    }

}



void
skcache_disk_setup()
{
    int i, j;
    long long tmp;
    int fdpoolsize = radix_h1*radix_h2;

    if(cache_index_prefix==NULL || *cache_index_prefix=='\0') {
        cache_index_prefix = cache_dir_prefix;
    }

    if(radix_disk<2) {
        for(i=0;i<nbucket;i++) {
            bk2fs[i] = 0;
        }
    }
    else {
        i = chk_bk2fs_map(0);
        Trace("chk fsmap ret %d\n", i);
        if(i) {
            Error("fsmap is wrong");
            fflush(stdout);
            fflush(stderr);
            /*
            show_bk2fs_map();
            */
            show_fs2bk_map();
            fflush(stdout);
            fflush(stderr);

            equlize_fsmap();

            i = chk_bk2fs_map(0);
            Trace("chk fsmap ret %d\n", i);
        }
    }

    fflush(stdout);
    fflush(stderr);
    show_fs2bk_map();
#if 0
    show_bk2fs_map();
#endif
    fflush(stderr);
    fflush(stdout);


    sim_stat_U_alloc_new(&swapin_time,  "swapin time [msec]",
        10, K(1),   swapin_time_count);
    sim_stat_U_alloc_new(&swapout_time,   "swapout time [msec]",
        10, K(1),   swapout_time_count);

    sim_stat_U_alloc_new(&checkdir_time,    "checkdir time [msec]",
        10, K(2),   checkdir_time_count);
    sim_stat_U_alloc_new(&checkdir_time_short, "checkdir time (short) [msec]",
        10, K(2),   checkdir_time_short_stat_count);

    sim_stat_alloc_new(&swapout_fail_reason,    "swapout fail reason",
        K(1),   swapout_fail_reason_count);
    sim_stat_alloc_new(&swapin_fail_reason,   "swapin fail reason",
        K(1),   swapin_fail_reason_count);

    sim_hist_alloc_new(&swapin_hist,    "swapin histroy [msec]",
        K(1), swapin_hist_area);
    sim_hist_alloc_new(&swapout_hist,   "swapout histroy [msec]",
        K(1), swapout_hist_area);

    if(radix_disk<=1) {
        j = 1;
    }
    else {
        j = radix_disk;
    }
    for(i=0;i<j;i++) {
        fs_swapin_count[i] = 0;
        fs_swapout_count[i] = 0;
        sim_hist_alloc_new(&fs_swapin_hist[i],  "fs swapin histroy [msec]",
            FS_HIST_LEN, fs_swapin_hist_area[i]);
        sim_hist_alloc_new(&fs_swapout_hist[i], "fs swapout histroy [msec]",
            FS_HIST_LEN, fs_swapout_hist_area[i]);
    }


    bfd_pool = (int*)MALLOC(sizeof(int)*fdpoolsize);
    ifd_pool = (int*)MALLOC(sizeof(int)*fdpoolsize);
    mlist_add_elem("index fds",(void*)ifd_pool, sizeof(int)*fdpoolsize);
    mlist_add_elem("packed fds",(void*)bfd_pool, sizeof(int)*fdpoolsize);

    ips_pool = (off_t*)MALLOC(sizeof(off_t)*fdpoolsize);
    bps_pool = (off_t*)MALLOC(sizeof(off_t)*fdpoolsize);
    mlist_add_elem("index poss",(void*)ips_pool, sizeof(off_t)*fdpoolsize);
    mlist_add_elem("packed poss",(void*)bps_pool, sizeof(off_t)*fdpoolsize);


    for(i=0;i<fdpoolsize;i++) {
        ifd_pool[i] = -1;
        bfd_pool[i] = -1;
        
        ips_pool[i] = (off_t)-1L;
        bps_pool[i] = (off_t)-1L;
    }

    dir_size_max    = OFILE_BLKSIZE*OFILE_NBLK_FILE;
    dir_size_hwater = dir_size_max*19/20;
    dir_size_lwater = dir_size_max*16/20;



    


    disk_sweep_interval = require_disk_sweep_interval;


    dir_bit = (u_int32_t*)MALLOC(sizeof(u_int32_t)*dir_num*DBIT_ELEM_WIDTH);
    mlist_add_elem("dir_bit", (void*)dir_bit,
        sizeof(u_int32_t)*dir_num*DBIT_ELEM_WIDTH);

    dir_map = (int*)MALLOC(sizeof(int)*fdpoolsize*OFILE_NBLK_FILE);
    mlist_add_elem("dir_map", dir_map, sizeof(int)*fdpoolsize*OFILE_NBLK_FILE);


    for(i=0;i<fdpoolsize;i++) {
        for(j=0;j<OFILE_NBLK_FILE;j++) {
            DMAP_SET(i,j,-1);
            DBIT_UNSET(i,j);
        }
    }



    /*
     * print parameters
     */
    Trace("\
WWW Object Caching Area upon Disk(Filesystem)s:\n\
    index               '%s'\n\
    body                '%s'\n\
\n",
        cache_index_prefix, cache_index_prefix);

    Trace("\
Object Packing:\n\
    OFILE_BLKSIZE       %7d\n\
    OFILE_NBLK_FILE     %7d\n\
    OFILE_NBLK_SEEK     %7d\n\
    OFILE_NBLK_OBJ      %7d range [0..%d)\n\
\n",
        OFILE_BLKSIZE, OFILE_NBLK_FILE, OFILE_NBLK_SEEK,
        OFILE_NBLK_OBJ, OFILE_BLKSIZE*OFILE_NBLK_OBJ);
    Trace("\
Disk:\n\
    %4d dirs  ; %d x %d\n\
    %4d files ; %d file pairs (index and body)\n\
              %8s %5s %10s %7s %14s %7s\n\
        INDEX %8ld %5ld %10lld %7.2f %14lld %7.2f\n\
        BODY  %8ld %5ld %10lld %7.2f %14lld %7.2f\n\
    Allocated %lld Bytes %.2f GB\n\
\n",
    dir_num, radix_h1, radix_h2,
    dir_num*2, dir_num,
    "block",  "xN", "single", "MB", "all", "GB",
    ESIZE, OFILE_NBLK_FILE, 
        (long long) ESIZE * OFILE_NBLK_FILE, 
        fdM(ESIZE * OFILE_NBLK_FILE), 
        (long long) dir_num * ESIZE * OFILE_NBLK_FILE, 
        fdG((long long) dir_num * ESIZE * OFILE_NBLK_FILE), 
    OFILE_BLKSIZE, OFILE_NBLK_FILE, 
        (long long) OFILE_BLKSIZE * OFILE_NBLK_FILE, 
        fdM( OFILE_BLKSIZE * OFILE_NBLK_FILE), 
        (long long) dir_num * OFILE_BLKSIZE * OFILE_NBLK_FILE, 
        fdG((long long) dir_num * OFILE_BLKSIZE * OFILE_NBLK_FILE),
    (long long)dir_num *
        (ESIZE+OFILE_BLKSIZE) * OFILE_NBLK_FILE,
        fdG((long long)dir_num *
            (ESIZE+OFILE_BLKSIZE) * OFILE_NBLK_FILE)
    );

    Trace("\n\
Bitmaps for directories:\n\
  SIZE_OF_UINT        %d\n\
  DBIT_BIT_WIDTH      %d\n\
  DBIT_VALUE_RANGE    %d\n\
  DBIT_VALUE_RES      %d\n\
  DBIT_ELEM_WIDTH     %d\n\
  DBIT_BYTE_WIDTH     %d\n",
        SIZE_OF_UINT,
        DBIT_BIT_WIDTH, DBIT_VALUE_RANGE, DBIT_VALUE_RES,
        DBIT_ELEM_WIDTH, DBIT_BYTE_WIDTH);
    Trace("\n");

    skcache_preload_files();

}

void
skcache_disk_init()
{
    TRACE_RECEIPT;

    clear_bk2fs_map();
}


